/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Auth extends Cert {
    /** Creates a new Auth from an Issuer, a Subject, a Tag,  
	a propagate boolean, and a Validity*/
    public Auth(SDSIPrincipal p, Subject s, Tag t, boolean prop, Validity v) {
	super(p,s,t,prop,v);
	int pExists = prop ? 1 : 0;
	SexpList propPart = new SexpList(new Sexp[] {
	    new SexpString("propagate")});
	Sexp [] issuerParts = new Sexp[2];
	issuerParts[0] = new SexpString("issuer");
	issuerParts[1] = p.getSrep();
	Sexp[] subjParts = new Sexp[2];
	subjParts[0] = new SexpString("subject");
	subjParts[1] = s.getSrep();
	Sexp[] authParts = new Sexp[4 + pExists + (v.isRestricted()? 1 : 0)];
	authParts[0] = new SexpString("cert");
	authParts[1] = new SexpList(issuerParts);
	authParts[2] = new SexpList(subjParts);
	if (prop) {
	    authParts[3] = propPart;
    }
	authParts[3+pExists] = t.getSrep();
    if (v.isRestricted()) {
      authParts[4+pExists] = v.getSrep();
    }
	srep = new SexpList(authParts);
    }
    public Auth(SexpList l, SDSIPrincipal issuer) throws SexpParseException {
	super(l);
	boolean delegateDefined, issuerLocDefined, subjectLocDefined;
	this.issuer = issuer;
	issuerLocDefined = subjectLocDefined = delegateDefined = false;
	Sexp[] certParts = l.arrayOfElements(); 
	for (int j = 1; j < certParts.length; j++) {
	    Sexp element = certParts[j];
	    if (element instanceof SexpList) {
		SexpList list = (SexpList) element;
		String label = list.getType();
		if (label.equals("tag")) {
		    if (t != null)
			throw new SexpParseException
			    ("More than one tag in cert");
		    else
			t = new Tag(list);
		}
		else if (label.equals("subject-info")) {
		    if (subjectLocDefined)
			throw new SexpParseException
			    ("More than one Subject info in cert");
		    else {
			subjectLocDefined = true;
			if (!list.allElementsStrings())
			    throw new SexpParseException
				("Subject info should contain only uri's");
		    }
		}
		else if (label.equals("issuer-info")) {
		    if (issuerLocDefined)
			throw new SexpParseException
			    ("More than one Issuer info in cert");
		    else {
			issuerLocDefined = true;
			if (!list.allElementsStrings())
			    throw new SexpParseException
				("Issuer info should contain only uri's");
		    }
		}   
		else if (label.equals("propagate")) {
		    if (delegateDefined)
			throw new SexpParseException
			    ("More than one Delegation info in cert");
		    else {
			delegateDefined = true;
			if (list.size() != 1)
			    throw new SexpParseException
				("Delegation should be only a single word");
			else
			    delegate = true;
		    } 
		}
		else parseItem(list);
	    }  
	    else
		throw new SexpParseException
		    ("Shouldn't have a non-list in a cert");
	}
	if (s == null || t == null) 
	    throw new SexpParseException("Need subject and tag in cert");
    }
    public String getType()
    {
	return "Authorization Certificate";
    }
    public String toShortString()
    {
	return issuer.toShortString()
	    + "[" + t.toShortString() + (delegate?"+":"") + "]"
	    + " -> " + s.toShortString();
    }

    public String getAbbreviation()
    {
	return "A";
    }
    public Tag getTag()
    {
	return t;
    }
    public boolean getDelegate()
    {
	return delegate;
    }
    public boolean reduces(Cert cert)
    {
	return (this.isReducing()
		&& (cert instanceof Auth)
		&& (cert.s instanceof SDSIPrincipal)
		&& this.issuer.samePrincipalAs((SDSIPrincipal)cert.s)
		&& cert.delegate);
    }
    public Cert reduce(Cert cert)
    {
	if (!this.reduces(cert)) return null;
	return new Auth(cert.issuer,
			this.s,
			reducedTag(cert.t),
			this.delegate,
			reducedValidity(cert.valid));
    }
}
