/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;
import java.util.*;

/**
   The basic class which represents certificates in SDSI/SPKI 2.0.
   @author Alexander Morcos, Sameer Ajmani
   @see Auth
   @see Def
   @see SDSIObject
**/
public abstract class Cert extends SDSIObject implements SequenceItem {
  SDSIPrincipal issuer;
  Subject s;
  Tag t;
  boolean delegate;
  Validity valid;
  boolean versionDefined,certDisplayDefined,commentDefined;

  public boolean sameAs(Object object)
  {
	if (object instanceof Cert) {
      Cert that = (Cert)object;
      return this.issuer.sameAs(that.issuer)
		&& this.s.sameAs(that.s)
		&& this.t.sameAs(that.t)
		&& (this.delegate == that.delegate);
	}
	return false;
  }

  /** Initializes the variables in the Cert, does
      not set up the SexpList representation.  This constructor
      is only called by subclasses which also do additional setup work*/
  protected Cert(SDSIPrincipal issuer, Subject s, Tag t, boolean delegate,
                 Validity valid) 
  {
	this.issuer = issuer;
	this.s = s;
	this.t = t;
	this.delegate = delegate;
	this.valid = valid;
	versionDefined=certDisplayDefined=commentDefined=false;
  }
  /** Initializes the variables in the Cert, including setting up the
      SexpList representation.  Also checks that the list is really
      a cert.  This constructor is only called by subclasses which also 
      do additional setup work.
      @exception SexpParseException if the SexpList is not a correctly
      formatted SDSI certificate.*/
  protected Cert(SexpList l) throws SexpParseException{
	super(l);
	issuer = null;
	s = null;
	t = null;
	delegate = false;
	valid = new Validity();
	versionDefined = certDisplayDefined = commentDefined = false;
	if (!l.getType().equals("cert")) 
      throw new SexpParseException("Not a certificate");
	//need to correct to check for other fields and except on bad fields
  }
  /** Creates a new Cert. Scans the SexpList to determine whether the Cert
      is a Authorization or a Name Definition and then it returns the
      newly created appropriate sub-class.
      @param l the SexpList that contains the certificate
      @return a Cert object that represents the certificate
      @exception SexpParseException if the SexpList is not a correctly
      formatted SDSI certificate*/
  public static Cert create(SexpList l) throws SexpParseException{
	Sexp[] certParts = l.arrayOfElements();
	SDSIObject issuerType = null;
	for (int j = 1; j < certParts.length; j++) {
      Sexp element = certParts[j];
      if (element instanceof SexpList) {
		SexpList list = (SexpList) element;
		String label = list.getType();
		if (label.equals("issuer")) {
          if (issuerType != null)
			throw new SexpParseException("More than one issuer "
                                         +"in cert");
          if (list.size() != 2
              || !(list.elementAt(1) instanceof SexpList))
			throw new SexpParseException("Issuer defined "
                                         +"incorrectly");
          issuerType = SDSIObject
			.principalParse((SexpList)list.elementAt(1), null);
          if (!(issuerType instanceof Name)
              && !(issuerType instanceof SDSIPrincipal)) {
			throw new SexpParseException("Issuer must be a name "
                                         +"or a principal");
          }
		}
      }
	}
	if (issuerType == null) {
      throw new SexpParseException("Must have an issuer in cert");
    }
	if (issuerType instanceof SDSIPrincipal) {
      return new Auth(l,(SDSIPrincipal)issuerType);
	} else if (issuerType instanceof Name) {
      return new Def(l,(Name)issuerType);
    } else {
      // assert: false
      throw new Error("should not get here");
    }
  }
  /** Parses a sub-list within the S-expression, this routine is called by
      the sub-class constructors.

      @exception SexpParseException if the sub-list does not correctly
      specify part of a Cert */
  protected void parseItem(SexpList list) throws SexpParseException {
	String label = list.getType();
	if (label.equals("subject")) {
      if (s != null)
		throw new SexpParseException("More than one subject in cert");
      else if (issuer == null)
		throw new SexpParseException("No issuer defined in cert");
      else {
		if (!(list.size() == 2)
		    || !(list.elementAt(1) instanceof SexpList))
          throw new SexpParseException("Subject of wrong form "
                                       +"in cert");
		SDSIObject potentialSubject =
          SDSIObject.principalParse((SexpList)list.elementAt(1),
                                    issuer);
		if (potentialSubject instanceof Subject)
          s = (Subject)potentialSubject;
		else 
          throw new SexpParseException("Subject is not one of the "
                                       +"allowed Subject objects");
      }
	}
	else if (label.equals("version")) {
      if (versionDefined)
		throw new SexpParseException("More than one version in cert");
      else {
		versionDefined = true;
		if (!(list.size() == 2)
		    || !(list.elementAt(1) instanceof SexpString))
          throw new SexpParseException("Version is of wrong form "
                                       +"in cert");
      }
	}
	else if (label.equals("display")) {
      if (certDisplayDefined)
		throw new SexpParseException("More than one display type "
                                     +"for cert");
      else {
		certDisplayDefined = true;
		if (!(list.size() == 2)
		    || !(list.elementAt(1) instanceof SexpString))
          throw new SexpParseException("Display type is of wrong "
                                       +"form in cert");
      }
	}
	else if (label.equals("valid")) {
      Enumeration elems = list.elements();
      elems.nextElement(); // skip type
      while (elems.hasMoreElements()) {
        Sexp elem = (Sexp)elems.nextElement();
        if (elem instanceof SexpList) {
          valid.update((SexpList)elem);
        } else {
          throw new SexpParseException("Validity is of wrong "
                                       +"form in cert");
        }
      }
	}
	else if (label.equals("comment")) {
      if (commentDefined)
		throw new SexpParseException("More than comment in cert");
      else {
		commentDefined = true;
		if (!(list.size() == 2)
		    || !(list.elementAt(1) instanceof SexpString))
          throw new SexpParseException("Comment is of wrong form "
                                       +"in cert");
      }
	}
	else if (label.equals("issuer"))
      ;//already checked for this
	else
      throw new SexpParseException("Unknown sub-type in cert: "+label);
  }
  /** Tests if the Cert is currently valid.
       
      @return <code>true</code> if the cert is valid now,
      <code>false</code> otherwise.  */
  public boolean valid()
  {
	return valid.valid();
  }

  /** Returns the issuer of this Cert.
      @return the SDSIPrincipal which is the issuer of this certificate.
  **/
  public SDSIPrincipal getIssuer()
  {
	return issuer;
  }

  /** Returns the subject of this Cert.  All SDSIObjects that can be
      subjects of certificates implement the Subject interface.

      @return the SDSIObject which is the subject of this certificate.
  **/
  public Subject getSubject()
  {
	return s;
  }

  public Validity getValidity()
  {
	return valid;
  }

  // returns the intersection of this certs validity with another validity
  protected Validity reducedValidity(Validity otherValid)
  {
	return this.valid.intersectWith(otherValid);
  }
  // returns the intersection of this certs tag with another tag
  protected Tag reducedTag(Tag otherTag)
  {
	return this.t.intersectWith(otherTag);
  }
  // Return true iff this is a "reducing" name certificate.
  // That is, it maps to a key, not a name.
  public boolean isReducing()
  {
	return (s instanceof SDSIPrincipal);
  }
  abstract public boolean reduces(Cert cert);
  abstract public Cert reduce(Cert cert);
}  
