/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import java.util.*;
import sdsi.control.*;
import sdsi.sexp.*;

/**
   Generates a Sequence of signed certificates that derives the given
   certificate.  Uses information on the certificate closure from the
   cache.
   @author Sameer Ajmani
**/
public class ClosureProver extends Utility {
    private Vector _chain = null;
    private Validity _valid = null;
    private Stack _stack = null;
    
    private void initialize(CertCache cache)
	throws SDSIException
    {
	if (DEBUG) System.out.println("initialize");
	if (cache == null)
	    throw new SDSIException("Arguments must not be null");
	// should this run the full closure here?
	_cache = cache;
	_chain = new Vector();
	_valid = new Validity();
	_stack = new Stack();
    }

    public Sequence prove(CertCache cache, Cert cert)
	throws SDSIException
    {
	if (DEBUG) System.out.println("prove (top-level)");
	if (anyNull(new Object[] {cache, cert}))
	    throw new SDSIException("Arguments must not be null");
	initialize(cache);
	if (proveCert(cert))
	    return getSequence(_chain);
	else
	    return null;
    }

    private boolean proveCert(Cert cert)
    {
	try {
	    if (anyNull(new Object[] {_cache, _chain, _stack, cert}))
		return false;

	    if (onStack(_stack, cert)) return false;
	    _stack.add(cert);

	    if (DEBUG) System.out.println(_stack.size()+" proveCert");

	    SDSIPublicKey issuer = getIssuerKey(cert);
	    SDSIPublicKey subject = getSubjectKey(cert);
	    SDSISignature sig = getIssuerSig(cert);
	    int oldSize = _chain.size();
	    try {
		if ((issuer != null) && (sig != null)) {
		    if (DEBUG) System.out.println(_stack.size()
						  +" Found signature");
		    if (!onChain(_chain, issuer)) {
			_chain.add(issuer);
			_chain.add(HashOp.create("md5"));
		    }

		    if ((subject != null) && !onChain(_chain, subject)) {
			_chain.add(subject);
			_chain.add(HashOp.create("md5"));
		    }

		    _chain.add(cert);
		    _chain.add(HashOp.create("md5"));
		    _chain.add(sig);
		    return true;
		}
	    } catch (SDSIException e) {
		_chain.setSize(oldSize);
	    }
	    
	    Set pairs = _cache.getReducingPairs(cert);
	    Iterator iter = pairs.iterator();
	    while (iter.hasNext()) {
		if (DEBUG) System.out.println(_stack.size()
					      +" Trying a pair...");
		CertCache.ReducingPair pair =
		    (CertCache.ReducingPair)iter.next();
		Cert reducee = pair.getReducee();
		Cert reducer = pair.getReducer();

		oldSize = _chain.size();
		if (proveCert(reducee) && proveCert(reducer)) {
		    return true;
		} else {
		    _chain.setSize(oldSize);
		}
	    }
	
	    return false;
	} finally {
	    // pop the stack
	    if (_stack.size() > 0)
		_stack.removeElementAt(_stack.size() - 1);
	}
    }

}
