/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Def extends Cert {
    Name issuerName;
    public boolean sameAs(Object object)
    {
	if (object instanceof Def) {
	    Def that = (Def)object;
	    return this.issuerName.sameAs(that.issuerName)
		&& super.sameAs(that);
	}
	return false;
    }

    /** MUST CHANGE BOTH CONSTRUCTORS SO WE CAN HAVE ONLY ONE ISSUER NAME*/
    /**Creates a new Def from an issuer name, a subject, and a Validity*/
    public Def(SDSIPrincipal issuer, String name, Subject s, Validity v)
	throws SDSIException
    {
	super(issuer,s,new Tag(),true,v);
	this.issuerName = new Name(issuer, new String[] {name});
	Sexp[] nameParts = new Sexp[2];
	nameParts[0] = new SexpString("issuer");
	nameParts[1] = issuerName.getSrep();
	Sexp[] subjParts = new Sexp[2];
	subjParts[0] = new SexpString("subject");
	subjParts[1] = s.getSrep();
	Sexp[] defParts = new Sexp[v.isRestricted()? 4 : 3];
	defParts[0] = new SexpString("cert");
	defParts[1] = new SexpList(nameParts);
	defParts[2] = new SexpList(subjParts);
    if (v.isRestricted()) {
      defParts[3] = v.getSrep();
    }
	srep = new SexpList(defParts);
    }
    
    /** Creates a new Def from a SexpList.  The Name which is the issuer is
	also given, because in practice, this constructor is called from
	within Cert.create which has already parsed the issuer out of the
	S-expression
	
	@exception SexpParseException if the SexpList does not correctly
	represent a SDSI name defintion certificate */
    public Def(SexpList l,Name issuerName) throws SexpParseException
    {
	super(l);
	delegate = true; //possibly need this here!!!!!!!!!!!
	if (issuerName.size() > 1)
	    throw new SexpParseException("Only one name can be used "+
					 "as an issuer");
	this.issuerName = issuerName;
	this.issuer = issuerName.getPrincipal();
	this.t = new Tag();
	Sexp[] certParts = l.arrayOfElements(); 
	for (int j = 1; j < certParts.length; j++) {
	    Sexp element = certParts[j];
	    if (element instanceof SexpList) {
		SexpList list = (SexpList) element;
		String label = list.getType();
		parseItem(list);
	    }
	    else
		throw new SexpParseException("Shouldn't have a non-list "
					     +"in a cert");
	}
	if (s == null) 
	    throw new SexpParseException("Need subject in cert");
    }
    public String getType()
    {
	return "Name Certificate";
    }
    public String toShortString()
    {
	return issuerName.toShortString() + " => " + s.toShortString();
    }
    public String getAbbreviation()
    {
	return "D";
    }
    public Name getName()
    {
	return issuerName;
    }
    public String getNameString()
    {
	return issuerName.getNameIndex(0);
    }
    // Return true iff this name cert reduces the given name cert.
    // That is, it maps the first name of given cert to a key.
    public boolean reduces(Cert cert)
    {
	return (this.isReducing()
		&& (cert.s instanceof Name)
		&& this.issuer.samePrincipalAs(((Name)cert.s).owner)
		&& this.issuerName.getNameIndex(0)
		.equalsIgnoreCase(((Name)cert.s).getNameIndex(0)));
    }
    // returns the subject of a definition after reduction by this cert
    // assumes this is a reducing cert
    private Subject reducedSubject(Name name) throws SDSIException
    {
	if (name.size() > 1) {
	    String[] names = new String[name.size()-1];
	    for (int i = 0; i < names.length; i++)
		names[i] = name.getNameIndex(i+1);
	    return new Name((SDSIPrincipal)this.s, names);
	} else {
	    return this.s;
	}
    }
    private Def reduceDef(Def def) throws SDSIException
    {
	return new Def(def.issuer,
		       def.issuerName.getNameIndex(0),
		       reducedSubject((Name)def.s),
		       reducedValidity(def.valid));
    }
    private Auth reduceAuth(Auth auth) throws SDSIException
    {
	return new Auth(auth.issuer,
			reducedSubject((Name)auth.s),
			auth.t,
			auth.delegate,
			reducedValidity(auth.valid));
    }
    // reduces returns the reduced version of the given cert, if possible
    // otherwise returns null
    public Cert reduce(Cert cert)
    {
	if (!this.reduces(cert)) return null;	
	try {
	    if (cert instanceof Def)
		return reduceDef((Def)cert);
	    else if (cert instanceof Auth)
		return reduceAuth((Auth)cert);
	    else
		return null;
	} catch (SDSIException e) {
	    return null;
	}
    }
}
