/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;
import java.security.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Hash extends SDSIPrincipal {
  /** Creates a new Hash from a S-expression list.
	@exception SexpParseException if the list is not a correctly
	formatted Hash. */
  public Hash(SexpList l) throws SexpParseException
  {
	super(l);
	if (!l.getType().equals("hash"))
      throw new SexpParseException("Not a hash");
	if ((l.size() < 3)
	    || !(l.elementAt(1) instanceof SexpString)
	    || !(l.elementAt(2) instanceof SexpString))
      throw new SexpParseException("Incorrectly formatted Hash object");
	String alg = this.getHashID();
	byte[] value = this.getHashValue();
	if (alg.equals("md5")) {
      if (value.length != 16)
		throw new SexpParseException("Wrong hash length for MD5");
	} else if (alg.equals("sha1")) {
      if (value.length != 20)
		throw new SexpParseException("Wrong hash length for SHA-1");
	} else 
      throw new SexpParseException("Unrecognized hash algorithm");
  }
  
  public Hash(String algorithm, byte[] value)
	throws NoSuchAlgorithmException, SexpParseException
  {
	this(new SexpList(new Sexp[] {
      new SexpString("hash"),
      new SexpString(algorithm),
      new SexpString(value) }));
  }

  public static Hash create(String algorithm, byte[] value)
	throws NoSuchAlgorithmException, SDSIException
  {
	try {
      return new Hash(algorithm, value);
	} catch (SexpParseException e) {
      throw new SDSIException("SexpParseException: "
                              + e.getMessage());
	}
  }

  // FIXME: replace with a factory method that calls the constructor
  public Hash(String algorithm, SDSIObject s)
	throws NoSuchAlgorithmException
  {
	byte[] hash;
	MessageDigest md;
	Sexp[] list = new Sexp[3];
	list[0] = new SexpString("hash");
	list[1] = new SexpString(algorithm);
	if (algorithm.equals("md5")) {
      md = MessageDigest.getInstance("MD5");
      hash = s.getHash(md);
      list[2] = new SexpString(hash);
	}
	else if (algorithm.equals("sha1")) {
      md = MessageDigest.getInstance("SHA-1");
      hash = s.getHash(md);
      list[2] = new SexpString(hash);
	}
	else
      throw new NoSuchAlgorithmException();
	srep = new SexpList(list);
  }
   
  public String getHashID()
  {
	// do we want getAlgID and share Key interface?
	// what else is in key interface
	return ((SexpString)srep.elementAt(1)).stringContent();
  }
  public byte[] getHashValue()
  {
	return ((SexpString)srep.elementAt(2)).bytesContent(); 
  }
  public String getType()
  {
	return "Hash";
  }
  public String getAbbreviation()
  {
	return "H";
  }
  public String toShortString()
  {
	String val = srep.elementAt(2).toReadableString(0,10,0);
	return "Hash " + val.substring(0, ((val.length() < 8) ?
                                       val.length() : 8));
  }
  // these override the implementation in SDSIObject
  public Hash getHash()
  {
	return this;
  }
  public boolean equals(Object object) {
	if (object instanceof Hash) {
      Hash that = (Hash)object;
      return this.getHashID().equals(that.getHashID())
		&& MessageDigest.isEqual(this.getHashValue(),
                                 that.getHashValue());
	}
	return false;
  }
  public int hashCode() {
	byte[] hash = this.getHashValue();
	int sum = 0;
	for (int i = 0; i < hash.length; i++)
      sum += hash[i]*hash[i];
	return sum;
  }
} 

