/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Name extends SDSIObject implements Subject{
  SDSIPrincipal owner;
  SexpString[] names;
    
  public boolean sameAs(Object object)
  {
	return (object instanceof Name)
      && this.sameNameAs((Name)object);
  }

  /**Creates a new name from an issuer and the list of names to follow*/
  public Name(SDSIPrincipal issuer, String[] words) throws SDSIException
  {
    if (words.length == 0) 
      throw new SDSIException("Name must contain at least one name");
    owner = issuer;
    names = new SexpString[words.length];
    for (int i=0;i<words.length;i++)
      names[i] = new SexpString(words[i]);
    validateSrep();
  }
    
  public Name(SexpList l, SDSIPrincipal issuer)
	throws SexpParseException, SDSIException
  {
	super(l);
	boolean principalIncluded;
	int size = l.size();
	if (!l.getType().equals("name"))
      throw new SexpParseException("Not a name");
	if (!(size >= 2))
      throw new SexpParseException("Incorrectly formatted name, "
                                   + "too few elements");
	if (l.elementAt(1) instanceof SexpList) {
      SDSIObject supposedPrincipal;
      if (!(size >=3)
          || !((supposedPrincipal = SDSIObject.principalParse
                ((SexpList)l.elementAt(1),null))
               instanceof SDSIPrincipal))
		throw new SexpParseException("Incorrectly formatted name");
      else {
		principalIncluded = true;
		owner = (SDSIPrincipal)supposedPrincipal;
      }
	}
	else {
      principalIncluded = false;
      if (issuer == null)
		throw new SDSIException("Attempt to create name "
                                + "without a principal");
      else
		owner = issuer;
	}
	int preNames = principalIncluded ? 2 : 1;
	names = new SexpString[size-preNames];
	for (int i = 0;i<size-preNames;i++) {
      if (l.elementAt(i+preNames) instanceof SexpString)
		names[i] = (SexpString)l.elementAt(i+preNames);
      else
		throw new SexpParseException("Names must be byte strings");
	}
  }
  public void expandNames(Name dontInclude, Name expansion)
  {
	int dontIncludeLength;
	if (dontInclude == null)
      dontIncludeLength = 0;
	else
      dontIncludeLength = dontInclude.names.length;
	SexpString[] newNames = new SexpString[names.length
                                          + expansion.names.length
                                          - dontIncludeLength];
	System.arraycopy(names,0,newNames,0,names.length);
	System.arraycopy(expansion.names, dontIncludeLength, newNames,
                     names.length, newNames.length - names.length);
	names = newNames;
	validateSrep();
  }
  public void replaceNames(Name oldPart, Name newPart,
                           SDSIPrincipal newOwner)
  {
	int replaceLength;
	if (newPart == null)
      replaceLength = 0;
	else
      replaceLength = newPart.names.length;
	SexpString[] newNames = new SexpString[names.length
                                          + replaceLength
                                          - oldPart.names.length];
	if (newPart != null)
      System.arraycopy(newPart.names,0,newNames,0,newPart.names.length);
	System.arraycopy(names, oldPart.names.length, newNames,
                     replaceLength, names.length - oldPart.names.length);
	names = newNames;
	owner = newOwner;
	validateSrep();
  }
  public boolean sameNamesAs(Name another)
  {
	if (another.names.length != names.length)
      return false;
	for (int i = 0;i<names.length;i++) {
      if (!another.names[i].equals(names[i]))
		return false;
	}
	return true;
  }
  public boolean prefixNamesOf(Name another)
  {
	if (another.names.length < names.length)
      return false;
	for (int i = 0;i<names.length;i++) {
      if (!another.names[i].equals(names[i]))
		return false;
	}
	return true;
  }
  public boolean sameNameAs(Name that)
  {
	return this.getPrincipal().samePrincipalAs(that.getPrincipal())
      && this.sameNamesAs(that);
  }
  public SDSIPrincipal getPrincipal()
  {
	return owner;
  }
  public String toShortString()
  {
	String first = owner.toShortString() + "'s "
      + names[0].toReadableString(0,20,0);
	if (names.length == 1)
      return first;
	if (names.length == 2)
      return first + "'s " + names[1].toReadableString(0,20,0);
	else // (names.length >= 3)
      return first + "'s ...";
  }
  private void validateSrep()
  {
	Sexp[] parts = new Sexp[names.length + 2];
	for (int i=0;i<names.length;i++)
      parts[i+2] = names[i];
	parts[0] = new SexpString("name");
	parts[1] = owner.getSrep();
	srep = new SexpList(parts);
  }
  public int size()
  {
	return names.length;
  }
  public String getAbbreviation()
  {
	return "N";
  }
  public String getNameIndex(int i)
  {
	if (i >= names.length)
      return null;
	else
      return names[i].toReadableString(0,1600,0);
  }
  public String[] getNames()
  {
	String[] result = new String[this.size()];
	for (int i = 0; i < result.length; i++)
      result[i] = getNameIndex(i);
	return result;
  }
}
