/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import java.security.*;
import sdsi.sexp.*;
import java.math.*;
import java.io.*;

/**
   Class to represent SDSI objects.
   The many subclasses are useful for providing specific 
   functionality inherent to individual types of objects.
   @author Alexander Morcos, Sameer Ajmani
**/
abstract public class SDSIObject implements Serializable{
    protected SexpList srep;
    protected transient	Hash _hash = null;
    protected transient	boolean _modified;
    public Hash getHash() throws NoSuchAlgorithmException
    {
	// subclasses must set _modified flag when they change srep
	if ((_hash == null) || _modified) {
	    _hash = new Hash("md5", this);
	    _modified = false;
	}
	return _hash;
    }
    
    public boolean sameAs(Object object)
    {
	return this.equals(object);
    }

    public boolean equals(Object object)
    {
	try {
	    // compare classes and hashes
	    return this.getClass().isInstance(object)
		&& this.getHash().equals(((SDSIObject)object).getHash());
	} catch (NoSuchAlgorithmException e) {
	    // FIXME: this is a silent failure!
	    return false;
	}
    }
    public int hashCode()
    {
	try {
	    // return hashCode of hash
	    return this.getHash().hashCode();
	} catch (NoSuchAlgorithmException e) {
	    // FIXME: this is a silent failure!
	    return 0;
	}
    }
    SDSIObject() {} 
    /**
       Creates a new SDSIObject from a SexpList.
       @param s the SexpList that represents the SDSIObject.
    */
    public SDSIObject(SexpList s)
    {
	srep = s;
    }

    /**
       Reads a SDSIObject from an input stream
       @param is the input stream to read from
       @param p the "creating principal" of the object
       @return the parsed SDSIObject
       @exception IOException error reading from stream
       @exception SDSIException error parsing object
    */
    static public SDSIObject readFrom(InputStream is, SDSIPrincipal p)
	throws IOException, SDSIException
    {
	try {
	    PushbackInputStream input = new PushbackInputStream(is);
	    Sexp sexp = Sexp.parse(input);
	    if (sexp instanceof SexpList)
          return SDSIObject.principalParse((SexpList)sexp, p);
	    else
          return null;
	} catch (SexpException e) {
	    throw new SDSIException("S-expression exception: "
				    +e.getMessage());
	} catch (SexpParseException e) {
	    throw new SDSIException("S-expression parse error: "
				    +e.getMessage());
	}
    }

    /**
       Reads a SDSIObject from an input stream, assumes creating
       principal is null.
       @param is the input stream to read from
       @return the parsed SDSIObject
       @exception IOException error reading from stream
       @exception SDSIException error parsing object
    **/
  static public SDSIObject readFrom(InputStream is)
    throws IOException, SDSIException
  {
    return readFrom(is, null);
  }
  

    /**
       Writes this SDSIObject to an output stream in readable form
       @param os the output stream to write to
       @exception IOException error writing to stream
    */
    public void writeReadable(OutputStream os) throws IOException
    {
	try {
	    os.write(this.toReadableString(72).getBytes("8859_1"));
	} catch (UnsupportedEncodingException e) {
	    throw new IOException("Unsupported character encoding");
	}
    }

    /**
       Writes this SDSIObject to an output stream in canonical form
       @param os the output stream to write to
       @exception IOException error writing to stream
    */
    public void writeCanonical(OutputStream os) throws IOException
    {
	os.write(this.srep.getCanonRep());
    }

    /**
       Generates the SDSI object that is represented by the SexpList in the
       argument.  Parses the SexpList in the argument and returns the
       appropriate subclass of SDSIObject that has been initialized by the
       SexpList.  Is capable of parsing Acl's, KeyHolder's, and Name's that
       may require a creating princpal to be known.

       @param l the SexpLIst to be parsed.
       @param p the SDSIPrincipal that should be treated
       as having created the object
       @return the SDSI objecet represented
       @exception SexpParseException if the SexpList cannot
       be succesfully parsed into a SDSIObject */

    public static SDSIObject principalParse(SexpList l, SDSIPrincipal p)
	throws SexpParseException
    {
	String type = l.getType();
	try {
	    if (type.equals("acl"))
		return new Acl(l,p);
	    if (type.equals("keyholder")) 
		return new KeyHolder(l,p);
	    if (type.equals("name"))
		return new Name(l,p);
	    else
		return SDSIObject.parse(l);
	} catch (SexpParseException e) {
	    throw new SexpParseException("Error parsing " + type
					 + ": " + e.getMessage(),l);
	} catch (SDSIException e) {
	    throw new SexpParseException("SDSI error: "
					 + e.getMessage(),l);//change
	}
    }
    /**
       Generates the SDSI object that is represented by the SexpList in the
       argument.  Parses the SexpList in the argument and returns the
       appropriate subclass of SDSIObject that has been initialized by the
       SexpList.

       @param l the SexpLIst to be parsed.
       @return the SDSI objecet represented
       @exception SexpParseException if the SexpList cannot be
       succesfully parsed into a SDSIObject */
    private static SDSIObject parse(SexpList l)
	throws SexpParseException, SDSIException
    {
	String type = l.getType();
	if (type.equals("cert"))
	    return Cert.create(l);
	if (type.equals("sequence"))
	    return new Sequence(l);
	if (type.equals("public-key"))
	    return SDSIPublicKey.create(l);
	if (type.equals("private-key"))
	    return SDSIPrivateKey.create(l);
	if (type.equals("secret-key"))
	    return new SDSISecretKey(l);
	if (type.equals("do"))
	    return Op.create(l);
	if (type.equals("signature"))
	    return new SDSISignature(l);
	if (type.equals("hash"))
	    return new Hash(l);
	if (type.equals("object-hash"))
	    return new ObjectHash(l);
	if (type.equals("tag"))
	    return new Tag(l);
	if (type.equals(Threshold.SDSI_TYPE))
	    return new Threshold(l);
	else
	    throw new SDSIException("Unrecognized object type: "+type);
    }
    /**
       Returns a short descriptive string representing this SDSIObject.
       @return a short string description of this SDSIObject.
    */
    abstract public String toShortString();

    /**
       Returns a single or double letter abbreviation this SDSIObject.
       @return an abbreviated description of this SDSIObject.
    */
    abstract public String getAbbreviation();

    /**
       Returns a string of this SDSIObject represented in canonical form.
       @return a string containing the canonical form of this SDSIObject.
    */
    public String toString()
    {
	return srep.toString();
    }
    /** 
	Returns a string representing this SDSIObject encoded as an
	S-Expression.  The String contains only printable characters, is
	nicely formatted, and is a valid S-Expression.

	@param width the maximum width of any line in the string.
	@return a nicely formatted string containing the S-Expression
	of this SDSIObject.  */
    public String toReadableString(int width)
    {
	return srep.toReadableString(0, width, 0);
    }

    /**
       Returns the hash of this SDSIObject.  Hashes the canonical
       representation of this SDSIObject and returns the resulting byte
       array.
       @param md the MessageDigest to use for the hash, it will be reset.
       @return the hash of the canonical representation of this SDSIObject
       as a byte array.  */
    public byte[] getHash(MessageDigest md)
    {
	md.reset();
	md.update(srep.getCanonRep());
	return md.digest();
    }
    /**
       Updates a signature with the canonical representation of this
       SDSIObject.
       
       @param s the signature to update.
       @exception SignatureException if the argument is not a properly
       initialized Signature object.  */
    public void sign(Signature s) throws SignatureException
    {
	// TODO: do I want to do this
	// or expose canonical rep idea somewhere, if so change hash also 
	s.update(srep.getCanonRep());
    }
    /**
       Returns a short string that represents the type of this SDSIObject.
       @return a string represent the type of SDSIObject we're dealing with.
    */
    public String getType()
    {
	return "SDSIObject";
    }
    /**
       Returns the S-Expression of this SDSIObject in canonical form as a
       SexpList

       @return the SexpList that encodes this SDSIObject in canonical
       form. */
    public SexpList getSrep()
    {
	return srep;
    }
    
}  
