/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;

/**
   The class which represents secret keys in SDSI.
   @author Alexander Morcos, Sameer Ajmani
**/
public class SDSISecretKey extends SDSIObject implements Subject{
  
    public String toShortString()
    {
	return "secret-key";
    }

    public String getAbbreviation()
    {
	return "s";
    }

 /** Creates a new SecretKey from a SexpList.
      @exception SexpParseException if the list is not a correctly
      formatted SDSI secret key.
 */
  public SDSISecretKey(SexpList l) throws SexpParseException{
    super(l);
    if (!l.getType().equals("secret-key"))
      throw new SexpParseException("Not a secret key");
    if ((l.size() < 2) || !(l.elementAt(1) instanceof SexpString))
      throw new SexpParseException("Incorrectly formatted SDSI Secret Key");
    String alg = this.getAlgID();
    String uri = this.getURI();
    boolean uriDefined = (uri == null) ? false : true;
    if (alg.equals("hmac-md5")) { //redo this later with String tokenizer
      boolean kDefined = false;
      Sexp[] keyParts = l.arrayOfElements();
      for (int i = 2; i < (uriDefined ? (keyParts.length-1) : keyParts.length); i++) {
        //check for neg numbers?  //store numbers?
        Sexp element = keyParts[i];
        if (element instanceof SexpList) {
          SexpList list = (SexpList) element;
          String label = list.getType();
          if (label.equals("k")) {
            if (kDefined)
              throw new SexpParseException("More than one key in HMAC Key");
            else {
              kDefined = true;
              if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
                throw new SexpParseException("Key encoded incorrectly in HMAC Key");
            }
          }
          else
            throw new SexpParseException("Unrecognized sub-type in HMAC Key: " + label);
        }     
        else  
          throw new SexpParseException("Shouldn't have a non-list other than a URI in a HMAC Key");
      }
      if (!kDefined)
        throw new SexpParseException("Must define key in HMAC Key");
    }
    else if (alg.equals("des-cbc-mac")) { //redo this later with String tokenizer
      boolean kDefined = false, nonceDefined = false;
      Sexp[] keyParts = l.arrayOfElements();
      for (int i = 2; i < (uriDefined ? (keyParts.length-1) : keyParts.length); i++) {
        //check for neg numbers?  //store numbers?
        Sexp element = keyParts[i];
        if (element instanceof SexpList) {
          SexpList list = (SexpList) element;
          String label = list.getType();
          if (label.equals("k")) {
            if (kDefined)
              throw new SexpParseException("More than one key in DES Key");
            else {
              kDefined = true;
              if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
                throw new SexpParseException("Key encoded incorrectly in DES Key");
            }
          }
          else if  (label.equals("nonce")) {
            if (nonceDefined)
              throw new SexpParseException("More than one nonce in DES Key");
            else {
              nonceDefined = true;
              if ((list.size() != 2) || !(list.elementAt(1) instanceof SexpString))
                throw new SexpParseException("Nonce encoded incorrectly in DES Key");
            }
          }
          else
            throw new SexpParseException("Unrecognized sub-type in DES Key: " + label);
        }     
        else  
          throw new SexpParseException("Shouldn't have a non-list other than a URI in a DES Key");
      }
      if (!kDefined || !nonceDefined)
        throw new SexpParseException("Must define key and nonce in DES Key");
    }
    else
      throw new SexpParseException("Unknown secret key algorithm");
  }
 
  private String getAlgID() {
    return ((SexpString)srep.elementAt(1)).stringContent();
  }
  
  /**
  Returns the URI of the SecretKey if there is one.
  @return a string that is the URI of the SecretKey if it has one, <code>null</code> otherwise.
  */
  public String getURI() { //here we assume uri is always at the end and any ending 
    int numElements = srep.size();
    Sexp lastElement = srep.elementAt(numElements - 1);
    if (lastElement instanceof SexpString)
      return ((SexpString)lastElement).stringContent();
    else
      return null;
  }
  public String getType() {
    return "Secret Key";
  }
}
