/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;
import java.security.*;
import java.math.*;
import java.util.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class SDSISignature extends SDSIObject implements SequenceItem {
    public SDSIPrincipal principal;
    public Hash hashOfObject;
  public String algo; // assumed to be same as public key's algo
  public byte[] sig;

    public boolean sameAs(Object object)
    {
	if (object instanceof SDSISignature) {
	    SDSISignature that = (SDSISignature)object;
	    return this.principal.sameAs(that.principal)
		&& this.hashOfObject.sameAs(that.hashOfObject);
	}
	return false;
    }
  
    public SDSISignature(SDSIObject signedObject, 
			 SDSIPrivateKey privKey, 
			 SDSIPublicKey pubKey) throws SignatureException{
	try {
	    this.principal = pubKey; 
	    this.hashOfObject = new Hash("md5",signedObject);
        this.algo = pubKey.getAlgorithm();
	    Signature privSign =
		Signature.getInstance(privKey.getAlgorithm());
	    privSign.initSign(privKey);
	    privSign.update(signedObject.getSrep().getCanonRep());
	    this.sig = privSign.sign();
	    Sexp[] sigParts = new Sexp[4];
	    sigParts[0] = new SexpString("signature");
	    sigParts[1] = hashOfObject.getSrep();
	    sigParts[2] = pubKey.getSrep();
	    sigParts[3] = new SexpList(new Sexp[] {
          new SexpString(algo), new SexpString(sig) });
	    srep = new SexpList(sigParts);
	}
	catch (NoSuchAlgorithmException e) {
	    throw new SignatureException(e.getMessage());
	}
	catch (InvalidKeyException e) {
	    throw new SignatureException(e.getMessage());
	}
    }
    public SDSISignature(SexpList l) throws SexpParseException{
	super(l);
	if (!l.getType().equals("signature"))
	    throw new SexpParseException("Not a signature");
	if ((l.size() < 4))
	    throw new SexpParseException("Incorrectly formatted "
					 +"Signature object: too few parts");
	if (!(srep.elementAt(1) instanceof SexpList)
	    || !(srep.elementAt(2) instanceof SexpList) 
	    || !(srep.elementAt(3) instanceof SexpList))
	    throw new SexpParseException("Incorrectly formatted "
					 +"Signature: parts of wrong type");
	SDSIObject supposedHash =
	    SDSIObject.principalParse((SexpList)srep.elementAt(1),null);
	if (!(supposedHash instanceof Hash))
	    throw new SexpParseException("Incorrectly formatted "
					 +"Signature: first part wasn't "
					 +"a hash");
	else
	    hashOfObject = (Hash)supposedHash;
	SDSIObject supposedPrincipal =
	    SDSIObject.principalParse((SexpList)srep.elementAt(2),null);
	if (!(supposedPrincipal instanceof SDSIPrincipal))
	    throw new SexpParseException("Incorrectly formatted "
					 +"Signature: principal must be "
					 +"Public Key or hash of one");
	else
	    principal = (SDSIPrincipal)supposedPrincipal;
    SexpList sigval = (SexpList)srep.elementAt(3);
    if (!(sigval.elementAt(0) instanceof SexpString)) {
	    throw new SexpParseException("Incorrectly formatted "
                                     +"Signature: signature algorithm "
                                     +"wasn't a string");
    } else {
      this.algo = ((SexpString)sigval.elementAt(0)).stringContent();
    }
    if (!(sigval.elementAt(1) instanceof SexpString)) {
	    throw new SexpParseException("Incorrectly formatted "
                                     +"Signature: signature value "
                                     +"wasn't a string");
    } else {
      this.sig = ((SexpString)sigval.elementAt(1)).bytesContent();
    }    
    }
    /*
      public void setSignedObject(SDSIObject o) {
      signedObject = o;
      }*/
    public String toShortString() {
	return "Signature";
    }
    public String getType() {
	return "Signature";
    }
    public String getAbbreviation()
    {
	return "S; ";
    }

    public SDSIPrincipal getPrincipal() {
	return principal;
    }
    /* 
       public SDSIObject getSignedObject() {
       return signedObject;
       }*/
    public Hash getHashOfObject() {
	return hashOfObject;
    }
    public boolean matches(SDSIObject object)
    {
	if (object == null) return false;
	try {
	    Hash hash = new Hash(this.hashOfObject.getHashID(), object);
	    return this.hashOfObject.equals(hash);
	} catch (NoSuchAlgorithmException e) {
	    return false;
	}
    }
    public byte[] getSignatureValue() { //need to worry about display type???
      return sig;
    }
    public boolean verify(SDSIPublicKey verifyKey, SDSIObject objToVerify)
	throws SexpParseException, NoSuchAlgorithmException,
	       SignatureException, InvalidKeyException
    {
	//watch for cast class exceptions when using SexpString/SexpList
	Signature verifySign =
	    Signature.getInstance(verifyKey.getAlgorithm());
	verifySign.initVerify(verifyKey);
	objToVerify.sign(verifySign);
	//Check keys are the same
	if (principal instanceof SDSIPublicKey) {
	    if (!(verifyKey.equals((SDSIPublicKey)principal)))
		throw new InvalidKeyException("Key given doesn't match key "
					      +"in signature");
	}
	else { //principal instanceof Hash
	    Hash hashPubKey = new Hash(((Hash)principal).getHashID(),verifyKey);
	    if (!((Hash)principal).equals(hashPubKey))
		throw new InvalidKeyException("Key given doesn't match key "
					      +"in signature");
	}
	//Checking Hash
	Hash HashObjToVerify = new Hash(hashOfObject.getHashID(),objToVerify);
	if (hashOfObject.equals(HashObjToVerify))
	    return verifySign.verify(sig);
	else
	    throw new SignatureException("Hash in signature doesn't match "
					 +"hash of object");
	//might want to treat this differently
    }
}
