/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
import java.io.*;
import java.util.*;

/**
   @author Sameer Ajmani
**/
public class Search {
    public static void main(String[] args) throws IOException
    {
        if (args.length < 1) {
            System.out.println("need cert filename");
            return;
        }
        StreamTokenizer in = new StreamTokenizer
            (new BufferedReader(new FileReader(args[0])));
        in.resetSyntax();
        in.eolIsSignificant(true);
        in.whitespaceChars('\u0000', '\u0020');
        in.wordChars('a','z');
        in.wordChars('A','Z');
        in.wordChars('0','9');
        in.wordChars('-','-');
        in.wordChars('>','>');
        in.commentChar('#');
        MultiMap map = new MultiMap();
        try {
            while (true) {
                parseCert(in, map);
            }
        } catch (EOFException e) {
            // ignored
        }
        BufferedReader stdin = new BufferedReader
            (new InputStreamReader(System.in));
        while (true) {
            System.out.print("Query: ");
            String query = stdin.readLine().trim();
            System.out.print("Target (enter for none): ");
            String target = stdin.readLine().trim();
            try {
                Set set = new HashSet();
                resolve(query, new StringBuffer(), target, map, set);
                System.out.println("Target not found.  Paths found:");
                Iterator paths = set.iterator();
                while (paths.hasNext()) {
                    System.out.println(paths.next().toString());
                }
            } catch (Exception e) {
                System.out.println("Target found.  Path:");
                System.out.println(e.getMessage());
            }
        }
    }

    private static void resolve(String query,
                                StringBuffer path,
                                String target,
                                MultiMap map,
                                Set set)
        throws Exception
    {
        boolean last = query.lastIndexOf(' ') == query.indexOf(' ');
        String subquery = query;
        while (isName(subquery)) {
            Iterator subjects = map.get(subquery).iterator();
            while (subjects.hasNext()) {
                String subject = (String)subjects.next();
                if (isName(subject)
                    && pathContains(path, certFor(subquery, subject))) {
                    continue;
                }
                StringBuffer extended = extend(path, subquery, subject);
                if (last) {
                    if (subject.equals(target)) {
                        throw new Exception(extended.toString());
                    }
                    if (!isName(subject)) {
                        set.add(extended.toString());
                    }
                }
                resolve(subject+query.substring(subquery.length()),
                        extended, target, map, set);
            }
            subquery = subquery.substring(0, subquery.lastIndexOf(' '));
        }
    }

    private static boolean pathContains(StringBuffer path, String cert) {
        return path.toString().indexOf(cert) >= 0;
    }
    
    private static boolean isName(String str) {
        return str.indexOf(' ') > 0;
    }
    
    private static StringBuffer extend(StringBuffer path,
                                       String query,
                                       String subject)
    {
        path = new StringBuffer(path.toString()); // clone
        return path.append('(').append(query)
            .append(" -> ").append(subject).append(')');
    }

    private static String certFor(String query, String subject)
    {
        return "("+query+" -> "+subject+")";
    }

    private static void parseCert(StreamTokenizer in, MultiMap map)
        throws IOException
    {
        String issuer = nextString(in);
        if (issuer == null) return;
        StringBuffer key = new StringBuffer(issuer);
        int inames = 0;
        while(true) {
            String iname = nextString(in);
            if (iname == null) return;
            if (iname.equals("->")) {
                break;
            }
            key.append(' ').append(iname);
            inames++;
        }
        if (inames < 1) {
            throw new IOException("no issuer names on line"
                                  +in.lineno());
        }
        String subject = nextString(in);
        if (subject == null) return;
        StringBuffer value = new StringBuffer(subject);
        try {
            while (true) {
                String sname = nextString(in);
                if (sname == null) return;
                value.append(' ').append(sname);
            }
        } finally {
            map.put(key.toString().trim(), value.toString().trim());
        }
    }

    private static String nextString(StreamTokenizer in)
        throws IOException
    {
        switch (in.nextToken()) {
        case StreamTokenizer.TT_WORD:
            return in.sval;
        case StreamTokenizer.TT_EOL:
            return null;
        case StreamTokenizer.TT_EOF:
            throw new EOFException();
        default:
            throw new IOException("unexpected character on line "
                                  +in.lineno());
        }
    }

    public static class MultiMap {
        Map map = new HashMap();
        void put(String key, String value) {
            Set set = (Set)map.get(key);
            if (set == null) {
                set = new HashSet();
            }
            set.add(value);
            System.out.println("adding ("+key+" -> "+value+")");
            map.put(key, set);
        }
        Set get(String key) {
            Set set = (Set)map.get(key);
            if (set == null) {
                return Collections.EMPTY_SET;
            }
            return Collections.unmodifiableSet(set);
        }
    }
}


