/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import java.util.*;
import sdsi.control.*;
import sdsi.sexp.*;

/**
   Finds the proof of a property in a CertCache without taking a closure.
   @author Sameer Ajmani
   @version 1.0
**/
public class SearchProver extends Utility {
    // holds the elements of the current proof
    private Stack _stack = null;

    private void initialize(CertCache cache)
        throws SDSIException
    {
        if (DEBUG) System.out.println("initialize");
        if (cache == null)
            throw new SDSIException("Arguments must not be null");
        _cache = cache;
        _stack = new Stack();
    }

    /**
       Looks in the given cache for a proof that the given issuer and names
       are bound to the target subject.
       @param cache the cache in which to search
       @param issuer the issuer of the first name
       @param names the names that bind to the subject
       @param target the target subject of the binding
       @return a proof sequence or null if no proof is found
       @exception SDSIException thrown if the arguments are null or there
       is an internal error
    */
    public Sequence proveDef(CertCache cache,
                             SDSIPrincipal issuer,
                             String[] names,
                             Subject target)
        throws SDSIException
    {
        if (DEBUG) System.out.println("proveDef (top-level)");
        if (anyNull(new Object[] {cache, issuer, names, target}))
            throw new SDSIException("Arguments must not be null");
        initialize(cache);
	
        if (proveName(new Name(issuer, names), target, _stack))
            return getSequence(_stack);
	
        return null;
    }

    /**
       Looks in the given cache for a proof that the tag can propagate from
       the issuer (with optional names) to the target subject.
       @param cache the cache in which to search
       @param issuer the issuer of the permission or of the first name
       @param names names for the issuer, may be null
       @param tag the permission tag
       @param target the target subject of the permission
       @return a proof sequence or null if no proof is found
       @exception SDSIException thrown if the arguments are null or there
       is an internal error */
    public Sequence proveAuth(CertCache cache,
                              SDSIPrincipal issuer,
                              String[] names, // may be null
                              Tag tag,
                              Subject target)
        throws SDSIException
    {
        if (DEBUG) System.out.println("proveAuth (top-level)");
        if (anyNull(new Object[] {cache, issuer, tag, target}))
            throw new SDSIException("Arguments must not be null");
        initialize(cache);
	
        if (proveAuth(issuer, names, tag, target, _stack))
            return getSequence(_stack);
 
        return null;
    }

    private boolean proveAuth(SDSIPrincipal issuer,
                              String[] names, // may be null
                              Tag tag,
                              Subject target,
                              Stack stack)
        throws SDSIException
    {
        if (DEBUG) System.out.println("proveAuth("
                                      +issuer.toShortString()+", names, "
                                      +tag.toShortString()+", "
                                      +target.toShortString()+", stack)");

        // issuer may be named; if so, resolve it first
        if (names != null)
            return proveAuth(new Name(issuer, names), tag, target, stack);
        else
            return proveAuth(issuer, tag, target, stack);
    }
    
    private boolean proveAuth(Subject issuer,
                              Tag tag,
                              Subject target,
                              Stack stack)
    {
        // dispatch according to subtype
        if (issuer instanceof SDSIPrincipal)
            return proveAuth((SDSIPrincipal)issuer, tag, target, stack);
	
        if (issuer instanceof Threshold)
            return proveThreshold((Threshold)issuer, tag, target, stack);

        if (issuer instanceof Name)
            return proveAuth(proveName((Name)issuer, new Stack()),
                             tag, target, stack);
	
        return false;
    }

    private boolean proveAuth(Set set,
                              Tag tag,
                              Subject target,
                              Stack stack)
    {
        if (DEBUG) System.out.println("proveAuth(set, "
                                      +tag.toShortString()+", "
                                      +target.toShortString()+", stack)");
        if ((stack == null) || (set == null)) {
            if (DEBUG) System.out.println("  FAIL: null parameter");
            return false;
        }
        Iterator pairs = set.iterator();
        int oldIndex = stack.size();
        // check if any of the issuers in the set leads to a proof
        while (pairs.hasNext()) {
            StackSubject pair = (StackSubject)pairs.next();
            // exclude non-principals (names resolved at lower level)
            if (!(pair._subject instanceof SDSIPrincipal)) continue;
            stack.addAll(pair._stack); // write the proof for the issuer
            if (proveAuth((SDSIPrincipal)pair._subject, tag, target, stack))
                return true;
            stack.setSize(oldIndex); // proof failed: restore the stack
        }

        if (DEBUG) System.out.println("  FAIL: no branch worked");
        return false;
    }

    private boolean proveAuth(SDSIPrincipal issuer,
                              Tag tag,
                              Subject target,
                              Stack stack)
    {
        if (DEBUG) System.out.println("proveAuth("
                                      +issuer.toShortString()+", "
                                      +tag.toShortString()+", "
                                      +target.toShortString()+", stack)");
        if (anyNull(new Object[] {_cache, issuer, tag, target}))
            return false;
        // try search from each matching auth in cache
        Iterator auths =
            sortCertsBySubject(_cache.getAuths(issuer, tag)).iterator();
        while (auths.hasNext()) {
            Auth auth = (Auth)auths.next();
            if (proveAuth(auth, target, stack)) return true;
        }
        return false;
    }

    private boolean proveAuth(Auth auth,
                              Subject target,
                              Stack stack)
    {
        if (DEBUG) System.out.println("proveAuth(auth, "
                                      +target.toShortString()+", stack)");
        if (anyNull(new Object[] {stack, auth, target})) return false;

        // avoid looping
        if (onStack(stack, auth)) {
            if (DEBUG) System.out.println("  FAIL: cert on stack");
            return false;
        }

        // add cert to proof
        int oldSize = stack.size();
        if (!pushCert(stack, auth)) {
            stack.setSize(oldSize);
            return false;
        }

        // continue proof
        if (proveAuthInner(auth, target, stack)) {
            return true;
        } else {
            // branch failed, restore stack
            stack.setSize(oldSize);
            return false;
        }
    }
    
    private boolean proveAuthInner(Auth auth,
                                   Subject target,
                                   Stack stack)
    {
        if (DEBUG) System.out.println("proveAuthInner(auth, "
                                      +target.toShortString()+", stack)");
        if ((auth == null) || (target == null)) return false;
	
        Subject subject = auth.getSubject();
	
        // check if subject is target
        if (subject.sameAs(target))
            return true;

        if (subject instanceof Name) {
            // check if the name binds to the target
            int oldSize = stack.size();
            if (proveName((Name)subject, target, stack))
                return true;
            stack.setSize(oldSize);
        }

        // delegate, if possible (FIXME: redundant name resolution!)
        if (auth.getDelegate())
            return proveAuth(subject, auth.getTag(), target, stack);

        return false;
    }

    private boolean proveThreshold(Threshold thresh,
                                   Tag tag,
                                   Subject target,
                                   Stack stack)
    {
        if (DEBUG) System.out.println("proveThreshold(thresh, "
                                      +tag.toShortString()+", "
                                      +target.toShortString()+", stack)");

        Subject[] subjects = thresh.getSubjects();
        int authorized = 0;

        // each branch of the threshold proof starts at the current stack
        Stack oldStack = (Stack)stack.clone();
        int oldIndex = stack.size();
        for (int i = 0; i < subjects.length; i++) {
            // restore original stack
            oldStack.setSize(oldIndex);

            // try to prove current subject
            if (proveAuth(subjects[i], tag, target, oldStack)) {
                // copy the new parts of the proof to the stack
                if (DEBUG) System.out.println("  branch "+i+" ok!");
                for (int j = oldIndex; j < oldStack.size(); j++)
                    pushStack(stack, (SDSIObject)oldStack.elementAt(j));
                authorized++;
                if (authorized >= thresh.getThreshold())
                    return true;
            } else {
                // this branch failed: tell verifier to skip it
                try {
                    if (DEBUG) System.out.println("  branch "+i+" failed:"
                                                  +" pushing skip op");
                    pushStack(stack, SkipOp.create());
                } catch (SDSIException e) {
                    // FIXME: Silent failure!
                    if (DEBUG) System.out.println("  FAIL: "+e);
                    return false;
                }
            }
        }
        if (DEBUG) System.out.println("  FAIL: could not satisfy threshold");
        return false;
    }

    private boolean proveName(Name name, Subject target, Stack stack)
    {
        if (DEBUG) System.out.println("proveName("
                                      +name.toShortString()+", "
                                      +target.toShortString()+", stack)");
        if ((name == null) || (target == null)) return false;

        // calculate all possible proof branches
        Set set = proveName(name, new Stack());
        if (set == null) return false;
        Iterator pairs = sortPairsByStackSize(set).iterator();
        while (pairs.hasNext()) {
            StackSubject pair = (StackSubject)pairs.next();
            if (pair._subject.sameAs(target)) {
                // return branch if finds target
                stack.addAll(pair._stack);
                return true;
            }
        }
        return false;
    }

    private Set proveName(Name name,
                          Stack stack)
    {
        if (DEBUG) System.out.println("proveName("+name.toShortString()
                                      +", stack)");
        if ((stack == null) || (name == null)) return null;
	
        String[] names = name.getNames();
        Set issuerSet = new HashSet();
        issuerSet.add(new StackSubject((Stack)stack.clone(),
                                       name.getPrincipal()));

        // resolve name component-by-component
        for (int i = 0; i < names.length; i++) {
            Iterator issuers = issuerSet.iterator();
            Set nextIssuerSet = new HashSet();
            // get potential subjects for each issuer
            while (issuers.hasNext()) {
                StackSubject pair = (StackSubject)issuers.next();
                // exclude non-principals (names resolved at lower level)
                if (!(pair._subject instanceof SDSIPrincipal)) continue;
                Set set = proveDef((SDSIPrincipal)pair._subject,
                                   names[i],
                                   pair._stack);
                if (set == null) continue;
                nextIssuerSet.addAll(set);
            }
            // results become new issuer set
            issuerSet = nextIssuerSet;
        }

        return issuerSet;	
    }

    private Set proveDef(SDSIPrincipal issuer,
                         String name,
                         Stack stack)
    {
        if (DEBUG) System.out.println("proveDef("+issuer.toShortString()
                                      +", "+name+", stack)");
        if (anyNull(new Object[] {_cache, issuer, name, stack}))
            return null;
        // search from each matching def in cache
        Iterator defs = _cache.getDefs(issuer, name).iterator();
        Set returnSet = new HashSet();
        while (defs.hasNext()) {
            Def def  = (Def)defs.next();
            Set set = proveDef(def, stack);
            if (set == null) continue;
            returnSet.addAll(set);
        }
        return returnSet;
    }

    private Set proveDef(Def def,
                         Stack stack)
    {
        if (DEBUG) System.out.println("proveDef(def, stack)");
        if (anyNull(new Object[] {def, stack})) return null;

        // avoid looping, but allow multiple reductions
        if (!def.isReducing() && onStack(stack, def)) {
            if (DEBUG) System.out.println("  FAIL: cert on stack");
            return null;
        }

        // add cert to proof
        int oldSize = stack.size();
        if (!pushCert(stack, def)) {
            stack.setSize(oldSize);
            return null;
        }

        // prove branch
        Set returnSet = proveDefInner(def, stack);

        // restore stack
        stack.setSize(oldSize);

        return returnSet;
    }

    private Set proveDefInner(Def def,
                              Stack stack)
    {
        if (DEBUG) System.out.println("proveDefInner(def, stack)");
        if ((def == null) || (stack == null)) return null;
	
        Subject subject = def.getSubject();
	
        Set returnSet = new HashSet();
        // return the subject
        returnSet.add(new StackSubject((Stack)stack.clone(),
                                       subject));

        // if the subject is a name, resolve it
        if (subject instanceof Name) {
            Set set = proveName((Name)subject, stack);
            if (set != null) returnSet.addAll(set);
        }
	
        return returnSet;
    }

    private boolean pushCert(Stack stack, Cert cert)
    {
        if (DEBUG) System.out.println("pushCert(stack, cert)");
        if ((_stack == null)
            || (stack == null)
            || (cert == null)) {
            if (DEBUG) System.out.println("  FAIL: null parameter");
            return false;
        }

        // check validity of _stack and stack together
        if (!validStack(_stack)
            .intersectWith(validStack(stack)
                           .intersectWith(cert.valid)).valid()) {
            if (DEBUG) System.out.println("  FAIL: cert invalidates proof");
            return false;
        }

        // get the cert's signature
        SDSISignature sig = getIssuerSig(cert);
        if (sig == null) {
            if (DEBUG) System.out.println("  FAIL: no issuer sig");
            return false;
        }

        // get the issuer's key
        SDSIPublicKey issuer = getIssuerKey(cert);
        if (issuer == null) {
            if (DEBUG) System.out.println("  FAIL: no issuer key");
            return false;
        }

        // get the subject's key (ok if null)
        SDSIPublicKey subject = getSubjectKey(cert);

        // write the keys, cert, and sig to the proof
        try {
            if (!onStack(_stack, issuer) && !onStack(stack, issuer)) {
                if (DEBUG) System.out.println("  pushing issuer key");
                pushStack(stack, issuer);
                pushStack(stack, HashOp.create("md5"));
            }
            if ((subject != null)
                && !onStack(_stack, subject) && !onStack(stack, subject)) {
                if (DEBUG) System.out.println("  pushing subject key");
                pushStack(stack, subject);
                pushStack(stack, HashOp.create("md5"));
            }
            pushStack(stack, cert);
            pushStack(stack, HashOp.create("md5"));
            pushStack(stack, sig);
        } catch (SDSIException e) {
            // FIXME: Silent failure!
            if (DEBUG) System.out.println("  FAIL: SDSI exception");
            return false;
        }
        return true;
    }

    private List sortCertsBySubject(Set set)
    {
        // assumes objects in set are Certs
        Class[] types = new Class[] {
            SDSIPrincipal.class, Name.class, Threshold.class };
        Vector[] buckets = new Vector[types.length+1];

        // initialize buckets
        for (int i = 0; i < buckets.length; i++)
            buckets[i] = new Vector();

        Iterator certs = set.iterator();
        while (certs.hasNext()) {
            Cert cert = (Cert)certs.next();
            // set index to appropriate bucket (types.length+1 ok)
            int index;
            for (index = 0; index < types.length; index++)
                if (types[index].isInstance(cert.getSubject()))
                    break;
            // put cert in bucket
            buckets[index].add(cert);
        }

        // output list in bucket order
        List list = new Vector();
        for (int i = 0; i < buckets.length; i++)
            for (int j = 0; j < buckets[i].size(); j++)
                list.add(buckets[i].elementAt(j));
        return list;
    }
    
    private List sortPairsByStackSize(Set set)
    {
        // assumes objects in set are StackSubjects
        Object[] array = set.toArray();
        Arrays.sort(array);
        // this could be replaced by Arrays.asList, but compiler bug
        List list = new Vector();
        for (int i = 0; i < array.length; i++)
            list.add(array[i]);
        return list;
    }

    private class StackSubject implements Comparable {
        public Stack _stack;
        public Subject _subject;
        public StackSubject(Stack stack, Subject subject)
        {
            _stack = stack;
            _subject = subject;
        }
        public boolean equals(Object object)
        {
            if (object instanceof StackSubject) {
                StackSubject that = (StackSubject)object;
                return this._stack.equals(that._stack)
                    && this._subject.equals(that._subject);
            }
            return false;
        }
        public int hashCode()
        {
            return _stack.hashCode() ^ _subject.hashCode();
        }
        public int compareTo(Object object)
        {   // can throw ClassCastException
            // order by stack length
            return this._stack.size()
                - ((StackSubject)object)._stack.size();
        }
    }
}
