/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;
import java.io.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Tag extends SDSIObject {
    TagExpression t; // may not want to carry this around
    // may just want to create it when needed and once at the beginning
    // to check formatting
    public Tag()
    {
	//the (*) tag with no S-exp representation
	t = new TagStar();
	srep = new SexpList(new Sexp[] { new SexpString("tag"),
					 new SexpList(new Sexp[] {
					     new SexpString("*")})}); 
    }
    public Tag(String string)
    {
	// a byte-string tag
	t = new TagByteString(new SexpString(string));
	srep = new SexpList(new Sexp[] { new SexpString("tag"),
					 new SexpString(string)});
    }
    public Tag(SexpList l) throws SexpParseException
    {
	super(l);
	if (l.size() != 2)
	    throw new SexpParseException("Incorrectly formatted tag");
	SexpList pTagStar;
	if (l.elementAt(1) instanceof SexpList 
	    && (pTagStar = ((SexpList)l.elementAt(1))).size() == 1 
	    && pTagStar.elementAt(0) instanceof SexpString 
	    && ((SexpString)pTagStar.elementAt(0)).stringContent().equals("*"))
	    t = new TagStar();
	else
	    t = TagExpression.parse(l.elementAt(1));
    }
    public String toShortString() {
	String s = srep.toReadableString(0,200,0);
	return s.substring(5,(s.length() < 21) ? s.length()-1 : 20);
    }
    public String getAbbreviation()
    {
	return "t";
    }

    public boolean intersect(Tag starFree) { //don't really like this
	TagExpression sf = starFree.t;
	return (t.intersect(sf) != null);
    }
    // intersect this tag with another tag
    public Tag intersectWith(Tag tag)
    {
	return tag; // need a sensible implementation!
    }
    public abstract static class TagExpression implements Serializable {
	public static TagExpression parse(Sexp s) throws SexpParseException
	{
	    if (s instanceof SexpString)
		return new TagByteString((SexpString)s);
	    else { //s is a SexpList
		SexpList l = (SexpList) s;
		if (l.elementAt(0) instanceof SexpList)
		    throw new SexpParseException
			("First element in tag can't be list");
		if (((SexpString)l.elementAt(0)).stringContent().equals("*")) {
		    return StarTag.parse(l);
		}
		else
		    return new SimpleTag(l);
	    }
	}
	public abstract TagExpression intersect(TagExpression starFree);
	public boolean equals(Object object)
	{
	    return (object instanceof TagExpression);
	}
    }
  
    public static class SimpleTag extends TagExpression {
	/*  so we don't allow tags that look like
	    ((* set read write) file foo) ?
	*/
	SexpString first;
	TagExpression[] rest;
	public boolean equals(Object object)
	{
	    if (object instanceof SimpleTag) {
		SimpleTag that = (SimpleTag)object;
		if (!this.first.equals(that.first)
		    || (this.rest.length != that.rest.length))
		    return false;
		for (int i = 0; i < this.rest.length; i++)
		    if (!this.rest[i].equals(that.rest[i]))
			return false;
		return super.equals(that);
	    }
	    return false;
	}
	public SimpleTag(SexpList l) throws SexpParseException
	{
	    int size = l.size();
	    first = (SexpString)l.elementAt(0);
	    rest = new TagExpression[size - 1];
	    for (int i =0;i<size-1;i++)
		rest[i] = TagExpression.parse(l.elementAt(i+1));
	}
	public TagExpression intersect(TagExpression starFree)
	{
	    if ((starFree instanceof StarTag)
		|| !(starFree instanceof SimpleTag)) return null;
	    SimpleTag simpTag = (SimpleTag)starFree;
	    if ((simpTag.rest.length != this.rest.length)
		|| !simpTag.first.equals(this.first)) return null;
	    for (int i = 0;i<this.rest.length;i++) {
		if ((simpTag.rest[i] = this.rest[i]
		     .intersect(simpTag.rest[i])) == null)
		    return null;
	    }
	    return simpTag;
	}
    }
  
    public abstract static class StarTag extends TagExpression{
	public boolean equals(Object object)
	{
	    return (object instanceof StarTag)
		&& super.equals(object);
	}
	public static StarTag parse(SexpList l) throws SexpParseException{
	    if (l.size() < 2 || !(l.elementAt(1) instanceof SexpString))
		throw new SexpParseException("Tag with star not encoded properly");
	    String starType = ((SexpString)l.elementAt(1)).stringContent();
	    if (starType.equals("set"))
		return new TagSet(l);
	    else if (starType.equals("range"))
		return new TagRange(l);
	    else if (starType.equals("prefix"))
		return new TagPrefix(l);
	    else
		throw new SexpParseException("Unknown star type in tag");
	}
    }
  
    public static class TagStar extends StarTag{
	public boolean equals(Object object)
	{
	    return (object instanceof TagStar)
		&& super.equals(object);
	}
	public TagStar() {;}
	public TagExpression intersect(TagExpression starFree) {
	    if (starFree instanceof StarTag)
		return null;
	    else 
		return starFree;
	}
    }
  
    public static class TagRange extends StarTag{
	boolean stringCompare = false;
	boolean strictlyGreater = false;
	boolean strictlyLess = false;
	SexpString lowerLimit = null;
	//can store these as strings if we're not carrying them everywhere
	SexpString upperLimit = null; //same in prefix
	public boolean equals(Object object)
	{
	    if (object instanceof TagRange) {
		TagRange that = (TagRange)object;
		return (this.stringCompare == that.stringCompare)
		    && (this.strictlyGreater == that.strictlyGreater)
		    && (this.strictlyLess == that.strictlyLess)
		    && this.lowerLimit.equals(that.lowerLimit)
		    && this.upperLimit.equals(that.upperLimit)
		    && super.equals(that);
	    }
	    return false;
	}
	public TagRange(SexpList l) throws SexpParseException { 
	    int size = l.size();
	    if (!(size == 3 || size ==5 || size == 7)) 
		throw new SexpParseException("*-range type tag not encoded properly");
	    for (int i=2;i<size;i++) {
		if (!(l.elementAt(i) instanceof SexpString))
		    throw new SexpParseException("*-range type tag not encoded properly");
	    }  
	    String rangeType = ((SexpString)l.elementAt(2)).stringContent();
	    if (rangeType.equals("alpha") || rangeType.equals("numeric") || rangeType.equals("time")
		|| rangeType.equals("binary") || rangeType.equals("date")) {
		stringCompare = true;
	    }    
	    else
		throw new SexpParseException("Unknown range type in tag");
	    if (size == 5) {
		String strictness = ((SexpString)l.elementAt(3)).stringContent();
		if (strictness.equals("l")) {
		    strictlyLess = true;
		    upperLimit = (SexpString)l.elementAt(4);
		}
		else if (strictness.equals("le")) {
		    strictlyLess = false;
		    upperLimit = (SexpString)l.elementAt(4);
		}
		else if (strictness.equals("g")) {
		    strictlyGreater = true;
		    lowerLimit = (SexpString)l.elementAt(4);
		}
		else if (strictness.equals("ge")) {
		    strictlyGreater = false;
		    lowerLimit = (SexpString)l.elementAt(4);
		}
		else 
		    throw new SexpParseException("*-range type tag not encoded properly");
	    }
	    if (size == 7) {
		String strictness = ((SexpString)l.elementAt(3)).stringContent();
		if (strictness.equals("g")) 
		    strictlyGreater = true;
		else if (strictness.equals("ge")) 
		    strictlyGreater = false;
		else 
		    throw new SexpParseException("*-range type tag not encoded properly");
		strictness = ((SexpString)l.elementAt(5)).stringContent();
		if (strictness.equals("l")) 
		    strictlyLess = true;
		else if (strictness.equals("le")) 
		    strictlyLess = false;
		else 
		    throw new SexpParseException("*-range type tag not encoded properly");
		lowerLimit = (SexpString)l.elementAt(4);
		upperLimit = (SexpString)l.elementAt(6);
	    }
	}
	public TagExpression intersect(TagExpression starFree) {
	    if (starFree instanceof StarTag)
		return null;
	    else {
		if (!(starFree instanceof TagByteString))
		    return null;
		SexpString starFreeString = ((TagByteString)starFree).value;
		try {
		    if (upperLimit != null) { //no upper and lower limit means all range
			if (strictlyLess) {
			    if (stringCompare)  //may want to move these??
				if (starFreeString.compareTo(upperLimit) >= 0)
				    return null;
				else
				    return null;
			}
			else {
			    if (stringCompare)
				if (starFreeString.compareTo(upperLimit) > 0)
				    return null;
				else
				    return null;
			}
		    }
		    if (lowerLimit != null) {
			if (strictlyGreater) {
			    if (stringCompare)
				if (starFreeString.compareTo(lowerLimit) <= 0)
				    return null;
				else
				    return null;
			}
			else {
			    if (stringCompare)
				if (starFreeString.compareTo(upperLimit) < 0)
				    return null;
				else 
				    return null;
			}
		    }
		    return starFree;
		}
		catch (SexpException e) {
		    return null;
		}
	    }
	}
    }
  
    public static class TagSet extends StarTag{
	TagExpression[] components;
	public boolean equals (Object object)
	{
	    if (object instanceof TagSet) {
		TagSet that = (TagSet)object;
		if (this.components.length != that.components.length)
		    return false;
		for (int i = 0; i < this.components.length; i++)
		    if (!this.components[i].equals(that.components[i]))
			return false;
		return super.equals(that);
	    }
	    return false;
	}
	public TagSet(SexpList l) throws SexpParseException{
	    int size = l.size();
	    components = new TagExpression[size - 2];
	    for (int i =0;i<size-2;i++)
		components[i] = TagExpression.parse(l.elementAt(i+2));
	}
	public TagExpression intersect(TagExpression starFree) {
	    if (starFree instanceof StarTag)
		return null;
	    else {
		for (int i = 0;i<components.length;i++) {
		    if (components[i].intersect(starFree) != null)  //this won't work unless intersection returns whole starfree
			return starFree;
		}
		return null;
	    }
	}
    }
  
    public static class TagPrefix extends StarTag {
	SexpString prefix;
	public boolean equals (Object object)
	{
	    if (object instanceof TagPrefix) {
		TagPrefix that = (TagPrefix)object;
		return this.prefix.equals(that.prefix)
		    && super.equals(that);
	    }
	    return false;
	}
	public TagPrefix(SexpList l) throws SexpParseException{
	    if (l.size() != 3 || !(l.elementAt(2) instanceof SexpString))
		throw new SexpParseException("Tag with *-prefix not encoded properly");
	    prefix = (SexpString)l.elementAt(2);
	}  
	public TagExpression intersect(TagExpression starFree) {
	    if (starFree instanceof StarTag)
		return null;
	    else {
		if (!(starFree instanceof TagByteString))
		    return null;
		SexpString starFreeString = ((TagByteString)starFree).value;
		if (starFreeString.startsWith(prefix))
		    return starFree;
		else
		    return null;
	    }
	}   
    }
  
    public static class TagByteString extends TagExpression {
	SexpString value;
	public boolean equals (Object object)
	{
	    if (object instanceof TagByteString) {
		TagByteString that = (TagByteString)object;
		return this.value.equals(that.value)
		    && super.equals(that);
	    }
	    return false;
	}
	public TagByteString(SexpString s) {
	    value = s;
	}
	public TagExpression intersect(TagExpression starFree) {
	    if (starFree instanceof StarTag)
		return null;
	    else {
		if (!(starFree instanceof TagByteString))
		    return null;
		if (this.value.equals(((TagByteString)starFree).value))
		    return starFree;
		else 
		    return null;
	    }
	}
    }
}
