/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;
import sdsi.*;
import sdsi.sexp.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Threshold extends SDSIObject implements Subject
{
    static final String SDSI_TYPE = "k-of-n";

    private int _threshold;
    private Subject[] _subjects;

    public String getType()
    {
	return "Threshold";
    }

    public String toShortString()
    {
	return "k-of-n ("+getThreshold()+"/"+getTotal()+")";
    }

    public int getThreshold()
    {
	return _threshold;
    }

    public int getTotal()
    {
	return _subjects.length;
    }

    public String getAbbreviation()
    {
	return "T";
    }

    public Subject[] getSubjects()
    {
	return _subjects;
    }

    public boolean sameAs(Object object)
    {
	if (object instanceof Threshold) {
	    Threshold that = (Threshold)object;
	    if (this._threshold != that._threshold) return false;
	    if (this._subjects.length != that._subjects.length) return false;
	    for (int i = 0; i < _subjects.length; i++)
		if (!this._subjects[i].sameAs(that._subjects[i]))
		    return false;
	    return true;
	}
	return false;
    }

    public static Threshold create(int threshold, Subject[] subjects)
	throws SDSIException
    {
	Sexp[] sexps = new Sexp[subjects.length+3];
	sexps[0] = new SexpString(Threshold.SDSI_TYPE);
	sexps[1] = new SexpString(Integer.toString(threshold));
	sexps[2] = new SexpString(Integer.toString(subjects.length));
	for (int i = 0; i < subjects.length; i++)
	    sexps[i+3] = ((SDSIObject)subjects[i]).getSrep();
	try {
	    return new Threshold(new SexpList(sexps));
	} catch (SexpParseException e) {
	    throw new SDSIException("SexpParseException: "+e.getMessage());
	}
    }

    Threshold(SexpList list) throws SexpParseException
    {
	super(list);
	if (list.size() < 3)
	    throw new SexpParseException("Threshold needs at least 3 items");
	if (!(list.getType().equals(Threshold.SDSI_TYPE)))
	    throw new SexpParseException("Threshold type expected: "
					 +list.getType());
	if (!(list.elementAt(1) instanceof SexpString))
	    throw new SexpParseException("SexpString expected in second slot");
	try {
	    _threshold = Integer.parseInt(((SexpString)list.elementAt(1))
					 .stringContent());
	    if (_threshold < 1)
		throw new SexpParseException("Threshold must be at least 1");
	} catch (NumberFormatException e) {
	    throw new SexpParseException("Integer expected in second slot");
	}
	if (!(list.elementAt(2) instanceof SexpString))
	    throw new SexpParseException("SexpString expected in third slot");
	int total;
	try {
	    total = Integer.parseInt(((SexpString)list.elementAt(2))
				     .stringContent());
	    if (total < _threshold)
		throw new SexpParseException("Total must be >= threshold");
	} catch (NumberFormatException e) {
	    throw new SexpParseException("Integer expected in third slot");
	}
	if (list.size() != (total+3))
	    throw new SexpParseException("Wrong number of elements in list: "
					 +list.size()+", expected: "
					 +(total+3));
	_subjects = new Subject[total];
	for (int i = 0; i < _subjects.length; i++) {
	    int listIndex = i+3;
	    if (!(list.elementAt(listIndex) instanceof SexpList))
		throw new SexpParseException("Expected list at index "
					     +listIndex);
	    try {
		SDSIObject object = SDSIObject.principalParse
		    ((SexpList)list.elementAt(listIndex), null);
		if (!(object instanceof Subject))
		    throw new SexpParseException("Expected Subject at index "
						 +listIndex);
		_subjects[i] = (Subject)object;
	    } catch (SexpParseException e) {
		throw new SexpParseException("Error parsing item "+listIndex
					     +": "+e.getMessage());
	    }
	}
    }
}
