/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import java.util.*;
import sdsi.control.*;
import sdsi.sexp.*;

/**
   @author Sameer Ajmani
**/
abstract public class Utility {
    final protected boolean DEBUG = true;
    protected CertCache _cache = null;

    protected boolean anyNull(Object[] objs)
    {
	if (objs == null)
	    throw new IllegalArgumentException();
	for (int i = 0; i < objs.length; i++)
	    if (objs[i] == null) {
		if (DEBUG) System.out.println("Object "+i+" is null");
		return true;
	    }
	return false;
    }

    protected SDSIPublicKey getKey(SDSIPrincipal principal)
    {
	return _cache.getKey(principal);
    }

    protected SDSIPublicKey getIssuerKey(Cert cert)
    {
	if ((cert == null) || (_cache == null)) return null;
	return getKey(cert.getIssuer());
    }

    protected SDSIPublicKey getSubjectKey(Cert cert)
    {
	Subject subject = cert.getSubject();
	if (subject instanceof SDSIPrincipal)
	    return getKey((SDSIPrincipal)subject);

	if (subject instanceof Name)
	    return getKey(((Name)subject).getPrincipal());

	return null;
    }

    protected SDSISignature getIssuerSig(Cert cert)
    {
	return _cache.getIssuerSig(cert);
    }

    protected Sequence getSequence(Vector vector) throws SDSIException
    {
	if (vector == null) return null;

	SequenceItem[] members = new SequenceItem[vector.size()];
	for (int i = 0; i < vector.size(); i++)
	    members[i] = (SequenceItem)vector.elementAt(i);

	try {
	    return Sequence.createFrom(members);
	} catch (SexpParseException e) {
	    throw new SDSIException("Internal Error: Bad Sequence");
	}
    }

    protected boolean onChain(Vector chain, SDSIObject object)
    {
	return (chain.indexOf(object) >= 0); // 0-based indexing
    }

    protected boolean onStack(Stack stack, SDSIObject object)
    {
	return (stack.search(object) > 0); // 1-based indexing
    }

    protected void pushStack(Stack stack, SDSIObject object)
    {
	stack.push(object);
    }

    protected void popStack(Stack stack)
    {
	if (!stack.empty())
	    stack.pop();
    }

    protected Validity validStack(Stack stack)
    {
	return getValidity(stack.toArray());
    }

    public Validity getValidity(Sequence sequence)
    {
	return getValidity(sequence.members());
    }

    private Validity getValidity(Object[] objects)
    {
	Validity valid = new Validity();
	for (int i = 0; i < objects.length; i++) {
	    if (objects[i] instanceof Cert)
		valid.update(((Cert)objects[i]).valid);
	}
	return valid;
    }
    
    protected SDSIObject topStack(Stack stack)
    {
	if (stack.empty())
	    return null;
	else
	    return (SDSIObject)stack.peek();
    }
}
