/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.sexp.*;
import java.text.*;
import java.util.*;
import java.io.*;

// FIXME: NOTHING IN HERE FOR ONLINE TESTS YET

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class Validity implements Serializable{
    static private SimpleDateFormat SDSIDate;
    Date notBeforeDate;
    Date notAfterDate;
    /*Note that dates are compared using Date objects and stored as such
      This is not in agreement with the spec which states byte string
      internal use I am using Dates because it is an easy way to check for
      the correct date format the only time there will be a difference is
      when we have a feb 30 kind of date then the ordering is not the same.
      Are those type of dates allowed?*/
    static {
	SDSIDate = new SimpleDateFormat("yyyy-MM-dd_HH:mm:ss");
	TimeZone utc = TimeZone.getTimeZone("GMT");
	SDSIDate.setTimeZone(utc);
    }
  
    public Validity() {
	notBeforeDate = null;
	notAfterDate = null;
    }

    public Validity intersectWith(Validity that)
    {
	Validity valid = new Validity();
	valid.update(this);
	valid.update(that);
	return valid;
    }

    public void updateStrings(String beforeString, String afterString)
	throws ParseException{
	Date bDate = null;
	Date aDate = null;
	if (!beforeString.equals(""))
	    updateBefore(getDate(beforeString));
	if (!afterString.equals(""))
	    updateAfter(getDate(afterString));
    }
  
    public void update(Validity other) {
	if (other.notBeforeDate != null)
	    updateBefore(other.notBeforeDate);
	if (other.notAfterDate != null)
	    updateAfter(other.notAfterDate);
    }
  
    public void updateBefore(Date updateDate) {
	if (notBeforeDate == null || updateDate.after(notBeforeDate))
	    notBeforeDate=updateDate;
    }
    public void updateAfter(Date updateDate) {
	if (notAfterDate == null || updateDate.before(notAfterDate))
	    notAfterDate=updateDate;
    }
  
    public void update(SexpList l) throws SexpParseException{
	if (l.getType().equals("not-before")) {
	    if (l.size() != 2)
		throw new SexpParseException("Date missing in valid field");
	    if (l.elementAt(1) instanceof SexpString) {
		Date updateDate = getDate((SexpString)l.elementAt(1));
		if (notBeforeDate == null || updateDate.after(notBeforeDate))
		    notBeforeDate=updateDate;
	    }
	    else
		throw new SexpParseException("Date not a string in "
					     +"valid field");
	}
	else if (l.getType().equals("not-after")) {
	    if (l.size() != 2)
		throw new SexpParseException("Date missing in valid field");
	    if (l.elementAt(1) instanceof SexpString) {
		Date updateDate = getDate((SexpString)l.elementAt(1));
		if (notAfterDate == null || updateDate.before(notAfterDate))
		    notAfterDate=updateDate;
	    }
	    else 
		throw new SexpParseException("Date not a string in "
					     +"valid field");
	}
	else if (l.getType().equals("online")) {
	} 
    }
  
    public static Date getDate(SexpString s) throws SexpParseException {
	try {
	    String notBefore = s.stringContent();
	    if (notBefore.length() != 19)
		throw new SexpParseException("Date wrong format in "
					     + "valid field");
	    Date result = SDSIDate.parse(notBefore);
	    return result;
	}
	catch (ParseException p) {
	    throw new SexpParseException("Date couldn't be parsed: "
					 + p.getMessage());
	}
    }
  
    public static Date getDate(String s) throws ParseException {
	try {
	    return SDSIDate.parse(s);
	} catch (ParseException e) {
	    throw e;
	} catch (Exception e) {
	    throw new ParseException(e.getMessage(),0);
	}
    }
  
  public boolean valid()
  {
	Date now = new Date();
	return ((notBeforeDate == null) || !now.before(notBeforeDate))
      &&  ((notAfterDate == null) ||  !now.after(notAfterDate));
  }
  // if this is false, this validity can be omitted from certs
  public boolean isRestricted()
  {
    return !(notBeforeDate == null && notAfterDate == null);
  }
  public Sexp getSrep()
  {
    ArrayList vList = new ArrayList();
    vList.add(new SexpString("valid"));
    if (notBeforeDate != null) {
      Sexp[] beforeParts = new Sexp[2];
      beforeParts[0] = new SexpString("not-before");
      beforeParts[1] = new SexpString(SDSIDate.format(notBeforeDate));
      vList.add(new SexpList(beforeParts));
    }
    if (notAfterDate != null) {
      Sexp[] afterParts = new Sexp[2];
      afterParts[0] = new SexpString("not-after");
      afterParts[1] = new SexpString(SDSIDate.format(notAfterDate));
      vList.add(new SexpList(afterParts));
    }
    Sexp[] vParts = new Sexp[vList.size()];
    vList.toArray(vParts);
    return new SexpList(vParts);
  }
}   
