/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi;

import sdsi.control.*;
import sdsi.sexp.*;

/**
   Verifies proof sequences for permissions and name bindings.
   @author Sameer Ajmani
   @version 1.0
**/
public class Verifier extends Utility {
    private SDSIObject[] _chain = null;  // chain being verified
    private int _index = 0;              // position in the chain
    private Validity _valid = null;      // validity to chain so far
    
    private void initialize(Sequence sequence)
        throws SDSIException
    {
        if (DEBUG) System.out.println("initialize");
        if (sequence == null)
            throw new SDSIException("Arguments must not be null");
        _cache = new CertCache();
        _chain = getSequenceArray(sequence);
        _index = 0;
        _valid = new Validity();
    }

    /**
       Checks if the given sequence proves that the tag propagates from the
       issuer (with optional names) to the target.
       @param sequence the proof sequence
       @param issuer the issuer of the permission or of the first name
       @param names names for the issuer, ok if null
       @param tag the permission tag
       @param target the target subject of the permission
       @return true iff the sequence proves the permission
       @exception SDSIException thrown if the arguments are null or there
       is an internal error */
    public boolean verifyAuth(Sequence sequence,
                              SDSIPublicKey issuer, // TODO: ok if principal?
                              String[] names, // null ok
                              Tag tag,
                              Subject target)
        throws SDSIException
    {
        if (DEBUG) System.out.println("verifyAuth (top-level)");
        if (anyNull(new Object[] {sequence, issuer, tag, target}))
            throw new SDSIException("Arguments must not be null");
        initialize(sequence);
        Subject subject;

        // if caller has provided names for the issuer, use them
        if (names != null)
            return verifyAuth(new Name(issuer, names), tag, target);
        else
            return verifyAuth(issuer, tag, target);
    }

    /**
       Checks if the given sequence proves that the issuer binds names to
       the target.
       @param sequence the proof sequence
       @param issuer the issuer of the first name
       @param names names to bind to the target
       @param target the target subject of the binding
       @return true iff the sequence proves the binding
       @exception SDSIException thrown if the arguments are null or there
       is an internal error */
    public boolean verifyDef(Sequence sequence,
                             SDSIPublicKey issuer, // TODO: ok if principal?
                             String[] names,
                             Subject target)
        throws SDSIException
    {
        if (DEBUG) System.out.println("verifyDef (top-level)");
        if (anyNull(new Object[] {sequence, issuer, names, target}))
            throw new SDSIException("Arguments must not be null");
        initialize(sequence);
        Subject subject = resolveName(new Name(issuer, names));
        return checkTarget(subject, target);
    }

    private boolean verifyAuth(Subject issuer,
                               Tag tag,
                               Subject target)
    {
        // dispatch according to subtype
        if (issuer instanceof SDSIPrincipal)
            return verifyAuth(getKey((SDSIPrincipal)issuer), tag, target);

        if (issuer instanceof Name)
            return verifyAuth(resolveName((Name)issuer), tag, target);

        if (issuer instanceof Threshold)
            return verifyThreshold((Threshold)issuer, tag, target);

        return false;
    }

    private boolean verifyAuth(SDSIPublicKey issuer,
                               Tag tag,
                               Subject target)
    {
        if (DEBUG) System.out.println("verifyAuth");
        if ((issuer == null) || (tag == null)) return false;

        // check certificate
        Cert cert = readCert(issuer);
        if ((cert == null)
            || !(cert instanceof Auth)
            || !cert.t.equals(tag)) // FIXME: use intersection of tags
            return false;
        
        // check if found target
        if (checkTarget(cert.s, target)) return true;

        // check if name resolves to target
        if (cert.s instanceof Name) {
            // attempt to resolve name
            // FIXME: This is an ambiguity in the grammar!
            //        We don't know whether to resolve the name
            //        or attmpt to propagate through it
            int oldIndex = _index;
            if (DEBUG) System.out.println("verifyAuth: resolving subject...");
            if (checkTarget(resolveName((Name)cert.s), target)) return true;
            if (DEBUG) System.out.println("verifyAuth: ...failed; propagate");
            _index = oldIndex; // restore stack pointer
        }       
        
        // delegate authority if possible
        if (((Auth)cert).getDelegate())
            return verifyAuth(cert.s, tag, target);

        return false;
    }

    private boolean verifyThreshold(Threshold thresh,
                                    Tag tag,
                                    Subject target)
    {
        if (DEBUG) System.out.println("verifyThreshold");
        if ((thresh == null) || (tag == null)) return false;

        Subject[] subjects = thresh.getSubjects();
        int authorized = 0;

        // attempt to verify permission for each subject
        for (int i = 0; i < subjects.length; i++) {
            if (readSkip()) continue; // skip unproven subjects

            if (verifyAuth(subjects[i], tag, target))
                authorized++;

            if (authorized >= thresh.getThreshold())
                return true;
        }
        return false;
    }

    private boolean readSkip()
    {
        if (_index >= _chain.length) return false;
        if (_chain[_index] instanceof SkipOp) {
            // read op (optional)
            if (DEBUG) System.out.println("readSkip: found SkipOp");
            _index++;
            return true;
        }
        return false;
    }

    private Subject resolveName(Name name)
    {
        if (DEBUG) System.out.println("resolveName");
        if (name == null) return null;

        Subject result = getKey(name.getPrincipal());
        String[] names = name.getNames();

        // resolve each component of the name
        for (int i = 0; i < names.length; i++) {
            if (!(result instanceof SDSIPrincipal)) return null;
            result = verifyDef(getKey((SDSIPrincipal)result), names[i]);
        }

        return result;
    }

    private Subject verifyDef(SDSIPublicKey issuer,
                              String name)
    {
        if (DEBUG) System.out.println("verifyDef");
        if ((issuer == null) || (name == null)) return null;

        // check certificate
        Cert cert = readCert(issuer);
        if ((cert == null)
            || !(cert instanceof Def)
            || !((Def)cert).getNameString().equals(name)) {
            if (DEBUG) System.out.println("verifyDef: FAIL: checks");
            return null;
        }
        
        // resolve subject if name
        if (cert.s instanceof Name) {
            // attempt to resolve name
            // FIXME: This is an ambiguity in the grammar!
            //        We don't know whether to resolve the name
            //        or return it directly (perhaps read a "stop"?)
            int oldIndex = _index;
            if (DEBUG) System.out.println("verifyDef: resolving subject...");
            Subject result = resolveName((Name)cert.s);
            if (result != null) return result;
            if (DEBUG) System.out.println("verifyDef: ...failed; return it");
            _index = oldIndex; // restore stack pointer
        }
        return cert.s;
    }

    private Cert readCert(SDSIPublicKey expectedIssuer)
    {
        if (DEBUG) System.out.println("readCert");
        if (anyNull(new Object[] {_cache, _chain, expectedIssuer, _valid}))
            return null;
        
        // read keys (optional)
        if (_index >= _chain.length) return null;
        while (_chain[_index] instanceof SDSIPublicKey) {
            if (DEBUG) System.out.println("readCert: key");
            _cache.add((SDSIPublicKey)_chain[_index]);
            _index++;
            if (_index >= _chain.length) return null;
            if (_chain[_index] instanceof HashOp) {
                // read op (optional)
                if (DEBUG) System.out.println("readCert: op1");
                _index++;  // we already did the hash
            }
            if (_index >= _chain.length) return null;
        }

        // read cert
        if (_index >= _chain.length) return null;
        if (!(_chain[_index] instanceof Cert)) return null;
        Cert cert = (Cert)_chain[_index];
        if (DEBUG) System.out.println("readCert: cert: "
                                      + cert.toShortString());
        _cache.add(cert);
        _index++;

        if (_index >= _chain.length) return null;
        if (_chain[_index] instanceof HashOp) {
            // read op (optional)
            if (DEBUG) System.out.println("readCert: op2");
            _index++;  // we already did the hash
        }

        // read signature
        if (_index >= _chain.length) return null;
        if (!(_chain[_index] instanceof SDSISignature)) return null;
        SDSISignature sig = (SDSISignature)_chain[_index];
        if (DEBUG) System.out.println("readCert: sig");
        _cache.add(sig);
        _index++;

        // check cert and signature
        try {
            SDSIPublicKey actualIssuer = getKey(cert.getIssuer());
            if ((actualIssuer == null)
                || !actualIssuer.equals(expectedIssuer)
                || !sig.verify(actualIssuer, cert)) {
                if (DEBUG) System.out.println
                               ("readCert: FAIL: sig checks:"
                                + (actualIssuer == null)
                                + !actualIssuer.equals(expectedIssuer)
                                + !sig.verify(actualIssuer, cert)
                                + "\nactual: "+actualIssuer.toShortString()
                                + ", expect: "+expectedIssuer.toShortString());
                return null;
            }

        } catch (SexpParseException e) {
            if (DEBUG) System.out.println("readCert: FAIL: "+e);
            return null;
        } catch (java.security.NoSuchAlgorithmException e) {
            if (DEBUG) System.out.println("readCert: FAIL: "+e);
            return null;
        } catch (java.security.SignatureException e) {
            if (DEBUG) System.out.println("readCert: FAIL: "+e);
            return null;
        } catch (java.security.InvalidKeyException e) {
            if (DEBUG) System.out.println("readCert: FAIL: "+e);
            return null;
        }

        // update validity
        _valid.update(cert.valid);
        if (!_valid.valid()) {
            if (DEBUG) System.out.println("readCert: FAIL: bad validity");
            return null;
        }
        
        return cert;
    }

    private boolean checkTarget(Subject actual,
                                Subject expected)
    {
        if (DEBUG) System.out.println("checkTarget");
        return ((actual != null)
                && (expected != null)
                && actual.sameAs(expected));
    }

    private SDSIObject[] getSequenceArray(Sequence sequence)
    {
        if (DEBUG) System.out.println("getSequenceArray");
        SDSIObject[] chain = new SDSIObject[sequence.members().length];
        System.arraycopy(sequence.members(), 0, chain, 0, chain.length);
        if (anyNull(chain)) return null;
        return chain;
    }

}
