/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi.control;

import java.util.*;
import java.io.*;
import java.security.*;
import sdsi.*;
import sdsi.sexp.*;

/**
   @author Sameer Ajmani
**/
public class CertCache implements Serializable{
  
    private Map hash2object = new HashMap();
    private MultiMap cert2pairs = new MultiMap();
    private MultiMap hash2sigs  = new MultiMap();
    private Map issuer2tag2auths = new HashMap();
    private Map issuer2name2defs = new HashMap();

    public CertCache()
    {
    }

    public CertCache(InputStream is) throws IOException
    {
	this();
	if (is == null) throw new IOException("Stream is null");
	try {
	    while (true) {
		SDSIObject obj = SDSIObject.readFrom(is);
		if (obj == null) break;
		this.add(obj);
	    }
	} catch (SDSIException e) {
	    throw new IOException("SDSIException: "+e.getMessage());
	}
    }

    synchronized public Set getObjects()
    {
	// since hash2object has multiple mappings, we need to use this
	// method to get a set that contains each object exactly once
	Set set = new HashSet();
	Iterator objects = hash2object.values().iterator();
	while (objects.hasNext())
	    set.add(objects.next());
	return set;
    }

    synchronized public void writeTo(OutputStream os) throws IOException
    {
	if (os == null) throw new IOException("Stream is null");
	Iterator objects = this.getObjects().iterator();
	while (objects.hasNext()) {
	    SDSIObject object = (SDSIObject)objects.next();
	    object.writeReadable(os);
	    os.write((byte)'\n');
	}
    }

    private void enqueuePairs(Vector queue, Cert cert)
    {
	SDSIObject object;
	Iterator objects = this.getObjects().iterator();
	while (objects.hasNext()) {
	    object = (SDSIObject)objects.next();
	    if (!(object instanceof Cert)) continue;
	    Cert other = (Cert)object;
	    if (cert.reduces(other))
		queue.add(new ReducingPair(cert, other));
	    if (other.reduces(cert))
		queue.add(new ReducingPair(other, cert));
	}
    }

    private void enqueueAllPairs(Vector queue)
    {
	SDSIObject object;
	Iterator reducers = this.getObjects().iterator();
	while (reducers.hasNext()) {
	    object = (SDSIObject)reducers.next();
	    if (!(object instanceof Cert)) continue;
	    Cert reducer = (Cert)object;
	    Iterator reducees = this.getObjects().iterator();
	    while (reducees.hasNext()) {
		object = (SDSIObject)reducees.next();
		if (!(object instanceof Cert)) continue;
		Cert reducee = (Cert)object;
		if (reducer.reduces(reducee))
		    queue.add(new ReducingPair(reducer, reducee));
	    }
	}
    }

    private void doClosure(Vector queue)
    {
	while (!queue.isEmpty()) {
	    ReducingPair pair = (ReducingPair)queue.elementAt(0);
	    queue.removeElementAt(0);
	    Cert cert = pair.reduce();
	    if ((cert != null)
		&& !getReducingPairs(cert).contains(pair)) {
		addReducingPair(cert, pair);
		enqueuePairs(queue, cert);
		add(cert);
	    }
	}
    }

    synchronized public void fullClosure()
    {
	Vector queue = new Vector();
	enqueueAllPairs(queue);
	doClosure(queue);
    }

    synchronized public void incrementalClosure(Cert cert)
    {
	if (cert == null) return;
	Vector queue = new Vector();
	enqueuePairs(queue, cert);
	doClosure(queue);
    }

    synchronized public Set getReducingPairs(Cert cert)
    {
	if (cert == null) return new HashSet();
	try {
	    return cert2pairs.get(cert.getHash());
	} catch (NoSuchAlgorithmException e) {
	    return null; // FIXME: silent error
	}
    }
    
    private void addReducingPair(Cert cert, ReducingPair pair)
    {
	Hash[] hashes = getAllHashes(cert);
	for (int i = 0; i < hashes.length; i++)
	    cert2pairs.put(hashes[i], pair);
    }

    // TODO: find a better way to maintain signature mappings
    synchronized public Set getSignatures(SDSIObject object)
    {
	if (object == null) return new HashSet();
	// signatures may be mapped under different hash functions,
	// so we need to get all signatures for each hash of object
	Hash[] hashes = getAllHashes(object);
	Set returnSet = new HashSet();
	for (int i = 0; i < hashes.length; i++)
	    if (hashes[i] != null)
		returnSet.addAll(hash2sigs.get(hashes[i]));
	return returnSet;
    }

    private SDSIObject getSignedObject(SDSISignature sig)
    {
	Hash hash = sig.getHashOfObject();
	SDSIObject signedObject = get(hash);
	if (!sig.matches(signedObject)) return null;
	return signedObject;
    }

    private void addSignature(SDSISignature sig)
    {
	// FIXME: Is there any reason to store signatures for non-certs?
	// And for certs, if there any reason to store non-issuer sigs?
	SDSIObject signedObject = getSignedObject(sig);
	if (signedObject != null) {
	    Hash[] hashes = getAllHashes(signedObject);
	    for (int i = 0; i < hashes.length; i++)
		hash2sigs.put(hashes[i], sig);
	} else {
	    // we have a sig with no matching object
	    hash2sigs.put(sig.getHashOfObject(), sig);
	}
    }

    private void removeSignature(SDSISignature sig)
    {
	SDSIObject signedObject = getSignedObject(sig);
	if (signedObject != null) {
	    Hash[] hashes = getAllHashes(signedObject);
	    for (int i = 0; i < hashes.length; i++)
		hash2sigs.remove(hashes[i], sig);
	} else {
	    // we have a sig with no matching object
	    hash2sigs.remove(sig.getHashOfObject(), sig);	    
	}
    }

    synchronized public Set getDefs(SDSIPrincipal issuer, String name)
    {
	try {
	    if ((issuer == null) || (name == null)) return new HashSet();
	    MultiMap name2defs =
		(MultiMap)issuer2name2defs.get(issuer.getHash());
	    if (name2defs == null)
		return new HashSet();
	    else
		return name2defs.get(name);
	} catch (NoSuchAlgorithmException e) {
	    throw new Error("No Provider (NoSuchAlgorithmException): "
			    +e.getMessage());
	}
    }

    private void addDef(Def def)
    {
	try {
	    MultiMap name2defs =
		(MultiMap)issuer2name2defs.get(def.getIssuer().getHash());
	    if (name2defs == null) name2defs = new MultiMap();
	    name2defs.put(def.getNameString(), def);
	    issuer2name2defs.put(def.getIssuer().getHash(), name2defs);
	} catch (NoSuchAlgorithmException e) {
	    throw new Error("No Provider (NoSuchAlgorithmException): "
			    +e.getMessage());
	}
    }

    private void removeDef(Def def)
    {
	try {
	    MultiMap name2defs =
		(MultiMap)issuer2name2defs.get(def.getIssuer().getHash());
	    if (name2defs == null) return;
	    name2defs.remove(def.getNameString(), def);
	    issuer2name2defs.put(def.getIssuer().getHash(), name2defs);
	    // is the last line necessary?
	} catch (NoSuchAlgorithmException e) {
	    throw new Error("No Provider (NoSuchAlgorithmException): "
			    +e.getMessage());
	}
    }

    synchronized public Set getAuths(SDSIPrincipal issuer, Tag tag)
    {
	try {
	    if ((issuer == null) || (tag == null)) return new HashSet();
	    MultiMap tag2auths =
		(MultiMap)issuer2tag2auths.get(issuer.getHash());
	    if (tag2auths == null)
		return new HashSet();
	    else
		return tag2auths.get(tag);
	} catch (NoSuchAlgorithmException e) {
	    throw new Error("No Provider (NoSuchAlgorithmException): "
			    +e.getMessage());
	}
    }

    private void addAuth(Auth auth)
    {
	try {
	    MultiMap tag2auths =
		(MultiMap)issuer2tag2auths.get(auth.getIssuer().getHash());
	    if (tag2auths == null) tag2auths = new MultiMap();
	    tag2auths.put(auth.getTag(), auth);
	    issuer2tag2auths.put(auth.getIssuer().getHash(), tag2auths);
	} catch (NoSuchAlgorithmException e) {
	    throw new Error("No Provider (NoSuchAlgorithmException): "
			    +e.getMessage());
	}
    }

    private void removeAuth(Auth auth)
    {
	try {
	    MultiMap tag2auths =
		(MultiMap)issuer2tag2auths.get(auth.getIssuer().getHash());
	    if (tag2auths == null) return;
	    tag2auths.remove(auth.getTag(), auth);
	    issuer2tag2auths.put(auth.getIssuer().getHash(), tag2auths);
	    // is the last line necessary?
	} catch (NoSuchAlgorithmException e) {
	    throw new Error("No Provider (NoSuchAlgorithmException): "
			    +e.getMessage());
	}
    }

    synchronized public SDSIObject get(Hash hash)
    {
	if (hash == null) return null;
	return (SDSIObject)hash2object.get(hash);
    }

    synchronized public void add(SDSIObject object)
    {
	if (object == null) return;
	addAllHashes(object);

	if (object instanceof Auth)
	    addAuth((Auth)object);
	else if (object instanceof Def)
	    addDef((Def)object);
	else if (object instanceof SDSISignature)
	    addSignature((SDSISignature)object);
    }

    private void addAllHashes(SDSIObject object)
    {
	Hash[] hashes = getAllHashes(object);
	for (int i = 0; i < hashes.length; i++)
	    if (hashes[i] != null)
		hash2object.put(hashes[i], object);
    }

    synchronized public void remove(SDSIObject object)
    {
	if (object == null) return;
	removeAllHashes(object);

	if (object instanceof Auth)
	    removeAuth((Auth)object);
	else if (object instanceof Def)
	    removeDef((Def)object);
	else if (object instanceof SDSISignature)
	    removeSignature((SDSISignature)object);
    }

    private void removeAllHashes(SDSIObject object)
    {
	Hash[] hashes = getAllHashes(object);
	for (int i = 0; i < hashes.length; i++)
	    if (hashes[i] != null)
		hash2object.remove(hashes[i]);
    }

    private static Hash[] getAllHashes(SDSIObject obj)
    {
	if (obj == null) return new Hash[] {};
	Hash[] hashes = new Hash[hashTypes.length];
	for (int i = 0; i < hashTypes.length; i++) {
	    try {
		hashes[i] = new Hash(hashTypes[i], obj);
	    } catch (NoSuchAlgorithmException e) {
		hashes[i] = null;
	    }
	}
	return hashes;
    }
  
    synchronized public SDSISignature getIssuerSig(Cert cert)
    {
	if (cert == null) return null;
	Iterator sigs = getSignatures(cert).iterator();
	while (sigs.hasNext()) {
	    SDSISignature sig = (SDSISignature)sigs.next();
	    if (sig.getPrincipal().sameAs(cert.getIssuer()))
		return sig;
	}
	return null;
    }

    synchronized public SDSIPublicKey getKey(Hash hash)
    {
	if (hash == null) return null;
	SDSIObject object = (SDSIObject)get(hash);
	if (object instanceof SDSIPublicKey)
	    return (SDSIPublicKey)object;
	return null;
    }

    synchronized public SDSIPublicKey getKey(SDSIPrincipal principal)
    {
    	if (principal instanceof SDSIPublicKey)
	    return (SDSIPublicKey)principal;

	if (principal instanceof Hash)
	    return getKey((Hash)principal);
	
	return null;
    }

    // FIXME: Maybe push all this into SDSIObject classes?
    //        Nice to have it centralized, though...
    synchronized public String getDisplayName(SDSIPrincipal owner, // null ok
					      SDSIObject o)
    {
	StringBuffer name = new StringBuffer();

	if (o == null) {
	    name.append("Unknown");
	} else if (o instanceof Hash) {
	    name.append("Hash " + ((Hash)o).getHashID()
			+ " " + getDisplayName(owner, get((Hash)o)));
	} else if (o instanceof ObjectHash) {
	    name.append("Obj "
			+ getDisplayName(owner,
					 ((ObjectHash)o).getHashOfObject()));
	} else if (o instanceof Name) {
	    Name n = (Name)o;
	    name.append(getNameFor(owner, n.getPrincipal()) + "'s "
			+ n.getNameIndex(0));
	    for (int i = 1; i < n.size(); i++) {
		name.append("'s " + n.getNameIndex(i));
	    }
	} else if (o instanceof Def) {
	    Def def = (Def)o;
	    // FIXME: this should use getIssuerSig(), but too slow:
	    // Change getIssuerSig to cache results, maybe?
	    name.append((getSignatures(def).isEmpty()?"":"* ") // signed?
			+ getNameFor(owner, def.getName().getPrincipal())
			+ "'s " + def.getNameString() + " => "
			+ getDisplayName(owner,
					 (SDSIObject)def.getSubject()));
	} else if (o instanceof Auth) {
	    Auth auth = (Auth)o;
	    // FIXME: this should use getIssuerSig(), but too slow:
	    // Change getIssuerSig to cache results, maybe?
	    name.append((getSignatures(auth).isEmpty()?"":"* ") // signed?
			+ getNameFor(owner, auth.getIssuer())
			+ " " + auth.getTag().toShortString()
			+ (auth.getDelegate()?"+":"") + " -> "
			+ getDisplayName(owner,
					 (SDSIObject)auth.getSubject()));
	} else if (o instanceof SDSIPublicKey) {
	    name.append("Key "+ getNameFor(owner, (SDSIPublicKey)o));
	} else if (o instanceof SDSISignature) {
	    name.append("Sig "
			+ getNameFor(owner, ((SDSISignature)o).principal));
	    SDSIObject signed = getSignedObject((SDSISignature)o);
	    name.append(" "+getDisplayName(owner, signed));
	} else if (o instanceof Sequence) {
	    name.append("Seq ");
	    SDSIObject[] members = ((Sequence)o).members();
	    for (int i = 0; i < members.length; i++)
		name.append((members[i] == null)?"?"
			    :members[i].getAbbreviation());
	} else if (o instanceof Op) {
	    Op op = (Op)o;
	    name.append("Op "+op.getOpcode());
	} else if (o instanceof Threshold) {
	    Threshold thresh = (Threshold)o;
	    name.append("Thresh("+thresh.getThreshold()
			+"/"+thresh.getTotal()+") ");
	    Subject[] subjects = thresh.getSubjects();
	    for (int i = 0; i < subjects.length; i++)
		name.append((subjects[i] == null)?"?"
			    :((SDSIObject)subjects[i]).getAbbreviation());
	} else {
	    name.append(o.toShortString());
	}
	return "["+name.toString()+"]";
    }

    synchronized public String getNameFor(SDSIPrincipal owner, // null ok
					  SDSIPrincipal principal)
    {
	if (principal == null) return "Unknown";
	if (principal.sameAs(owner)) return "SELF";
	SDSIPublicKey key = getKey(principal);
	if (key == null) return "Unknown";
	return key.toShortString();
    }

    private static final String[] hashTypes = new String[] {"md5","sha1"};

    // inner classes
    public class ReducingPair {
	private Cert _reducer;
	private Cert _reducee;
	ReducingPair(Cert reducer, Cert reducee)
	{
	    _reducer = reducer;
	    _reducee = reducee;
	}
	Cert reduce()
	{
	    return _reducer.reduce(_reducee);
	}
	public int hashCode()
	{
	    return (_reducer.hashCode() ^ _reducee.hashCode());
	}
	public boolean equals(Object object)
	{
	    return (object instanceof ReducingPair)
		&& ((ReducingPair)object)._reducer.equals(this._reducer)
		&& ((ReducingPair)object)._reducee.equals(this._reducee);
	}
	public Cert getReducer()
	{
	    return _reducer;
	}
	public Cert getReducee()
	{
	    return _reducee;
	}
    }

    private class MultiMap {
	Map _map = new HashMap();
	private Set getInternal(Object key)
	{
	    Object o = _map.get(key);
	    if (!(o instanceof Set)) {
		Set newSet = new HashSet();
		_map.put(key, newSet);
		return newSet;
	    }
	    return (Set)o;
	}
	public Set get(Object key)
	{
	    return Collections.unmodifiableSet(getInternal(key));
	}
	public void put(Object key, Object object)
	{
	    Set set = this.getInternal(key);
	    set.add(object);
	    _map.put(key, set);
	}

	public void remove(Object key, Object object)
	{
	    Set set = this.getInternal(key);
	    set.remove(object);
	    _map.put(key, set);
	}
    }
}
