/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package sdsi.control;

import sdsi.sexp.*;
import sdsi.*;
import java.awt.GridBagConstraints;
import java.awt.CardLayout;
import java.awt.Font;
import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.*;
import java.util.*;
import java.security.*;
import java.security.interfaces.*;
import java.math.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;

/**
   @author Alexander Morcos, Sameer Ajmani
**/
public class SDSIMainFrame extends JFrame
  implements ActionListener, ItemListener
{

  CertCache cache;
  SDSIObject currentObject;       // TODO: do we really need this?
  SDSIPublicKey currentPublicKey; // TODO: put these into a key pair
  SDSIPrivateKey currentPrivateKey;
  // list of current objects
  DefaultListModel currentObjects = new DefaultListModel();
  // for sequence creation
  DefaultListModel currentSequence = new DefaultListModel();
  // for threshold creation
  DefaultListModel currentSubjects = new DefaultListModel();

  public static void main(String argv[]) {
	Security.addProvider(new cryptix.provider.Cryptix());
	SDSIMainFrame base = new SDSIMainFrame();
	base.setVisible(true);
  }

  private class DisplayObject implements Comparable
  {
	String _name;
	SDSIObject _object;
	DisplayObject(SDSIObject object)
	{
      if (object == null) _name = "null";  // XXX needed?
      else
		_name = cache.getDisplayName(currentPublicKey, object);
      _object = object;
	}
	String getName()
	{
      return _name;
	}
	SDSIObject getObject()
	{
      return _object;
	}
	public String toString()
	{
      return _name;
	}
	public boolean equals(Object other)
	{
      if (other instanceof DisplayObject) {
		DisplayObject that = (DisplayObject)other;
		return this._name.equals(that._name)
          && this._object.equals(that._object);
      }
      return false;
	}
	public int hashCode()
	{
      return _name.hashCode() ^ _object.hashCode();
	}
	public int compareTo(Object other)
	{
      // can throw ClassCastException
      return this._name.compareTo(((DisplayObject)other)._name);
	}
  }

  private GridBagConstraints gbc = new GridBagConstraints();
  private void setGBC(int gridx, int gridy,
                      double weightx, double weighty)
  {
	gbc.gridx = gridx;
	gbc.gridy = gridy;
	gbc.weightx = weightx;
	gbc.weighty = weighty;
  }
  private void setGBC(int anchor, int fill)
  {
	gbc.anchor = anchor;
	gbc.fill = fill;
  }

  private void initializeCache()
  {
	this.cache = new CertCache();
	this.currentObject = null;
  }

  public SDSIMainFrame()
  {
	initializeCache();
	this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	
	//{{INIT_CONTROLS
	this.setTitle("SDSI Certificate Utility");
	this.getContentPane().setFont(new Font("Dialog",Font.PLAIN,14));
	JPanel mainPanel = new JPanel();
    mainPanel.setLayout(new BorderLayout(0, 0));
	this.getContentPane().add(mainPanel);
	// the main window
	Box contentBox = Box.createHorizontalBox();
	mainPanel.add("Center", contentBox);
	Box displayBox = Box.createVerticalBox();
	contentBox.add(displayBox);
	Box rightBox = Box.createVerticalBox();
	contentBox.add(rightBox);
    Box bottomBox = Box.createVerticalBox();
    mainPanel.add("South", bottomBox);
	// status bar
	Box statusBox = Box.createHorizontalBox();
	bottomBox.add(statusBox);
	statusBox.add(new JLabel("Status: ", JLabel.RIGHT));
	statusText = new JTextField("Initialized", 40);
	statusText.setEditable(false);
	statusBox.add(statusText);
	// identity bar
	Box identityBox = Box.createHorizontalBox();
	bottomBox.add(identityBox);
	identityBox.add(new JLabel("Identity: ", JLabel.RIGHT));
	identityText = new JTextField("None (Load Identity)", 40);
	identityText.setEditable(false);
	identityBox.add(identityText);

	// the display for the selected object
	displayBox.add(new JLabel("SDSI Object"));
	display = new JTextArea("", 25, 80);
	display.setFont(new Font("Monospaced",Font.PLAIN, 14));
	display.setEditable(false);
	displayBox.add(new JScrollPane(display));

	// the object list on the upper right
	panelObjList = new JPanel();
	rightBox.add(panelObjList);
	panelObjList.setLayout(new BorderLayout(0, 0));
	JLabel objListLabel = new JLabel("Current Objects");
	panelObjList.add("North", objListLabel);
	objectList = new JList(currentObjects);
	panelObjList.add("Center", new JScrollPane(objectList));
	
	// the control space on the lower right
	controls = new JTabbedPane();
	rightBox.add(controls);

	// control set: create object
	panelObjects = new JPanel();
	controls.addTab("Object", panelObjects);
	panelObjects.setLayout(new GridBagLayout());
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.NONE);
	setGBC(0,0,1,0);
	panelObjects.add(new JLabel("Create Object"), gbc);
	newObject = new JTextArea(12, 40);
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.BOTH);
	setGBC(0,1,1,1);
	panelObjects.add(newObject, gbc);
	createObject = new JButton("Create");
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.NONE);
	setGBC(0,2,1,0);
	panelObjects.add(createObject, gbc);

	// control set: create name
	panelName = new JPanel();
	controls.addTab("Name", panelName);
	panelName.setLayout(new GridLayout(4,1));
	panelName.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(new JLabel("Create Name"));
	panelName.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(issuerNButton = new JButton("Issuer Key"));
	panelTemp.add(issuerNSelect = new JTextField(25));
	issuerNSelect.setEditable(false);
	panelName.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(new JLabel("Names:"));
	panelTemp.add(nameText = new JTextField(30));
	nameText.setEditable(true);
	panelName.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(createNameButton = new JButton("Create"));

	// control set: create sequence
	panelSequence = new JPanel();
	controls.addTab("Sequence", panelSequence);
	panelSequence.setLayout(new GridBagLayout());
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.NONE);
	setGBC(0,0,1,0);
	panelSequence.add(new JLabel("Create Sequence"), gbc);
	sequenceList = new JList(currentSequence);
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.BOTH);
	setGBC(0,1,1,1);
	panelSequence.add(new JScrollPane(sequenceList), gbc);
	panelTemp = new JPanel();
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(addSeqButton = new JButton("Add"));
	panelTemp.add(insertSeqButton = new JButton("Insert"));
	panelTemp.add(removeSeqButton = new JButton("Remove"));
	panelTemp.add(clearSeqButton = new JButton("Clear"));
	panelTemp.add(createSeqButton = new JButton("Create"));
	panelTemp.add(loadSeqButton = new JButton("Load"));
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.NONE);
	setGBC(0,2,1,0);
	panelSequence.add(panelTemp, gbc);

	// control set: create threshold
	panelThreshold = new JPanel();
	controls.addTab("Threshold", panelThreshold);
	panelThreshold.setLayout(new GridBagLayout());
  	panelTemp = new JPanel();
  	panelTemp.add(new JLabel("Create Threshold:"));
  	panelTemp.add(thresholdText = new JTextField(3));
  	setGBC(GridBagConstraints.CENTER, GridBagConstraints.NONE);
  	setGBC(0,0,1,0);
  	panelThreshold.add(panelTemp, gbc);
	subjectsList = new JList(currentSubjects);
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.BOTH);
	setGBC(0,1,1,1);
	panelThreshold.add(new JScrollPane(subjectsList), gbc);
	panelTemp = new JPanel();
	panelTemp.add(addSubButton = new JButton("Add"));
	panelTemp.add(insertSubButton = new JButton("Insert"));
	panelTemp.add(removeSubButton = new JButton("Remove"));
	panelTemp.add(clearSubButton = new JButton("Clear"));
	panelTemp.add(createSubButton = new JButton("Create"));
	panelTemp.add(loadSubButton = new JButton("Load"));
	setGBC(GridBagConstraints.CENTER, GridBagConstraints.NONE);
	setGBC(0,2,1,0);
	panelThreshold.add(panelTemp, gbc);

	// control set: create hash
	panelHash = new JPanel();
	controls.addTab("Hash", panelHash);
	panelHash.setLayout(new GridLayout(2,1));

	panelHash.add(panelTemp = new JPanel());	
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(new JLabel("Item to Hash:"));
	panelTemp.add(hashItemName = new JTextField(25));
	hashItemName.setEditable(false);

	panelHash.add(panelTemp = new JPanel());	
	panelTemp.setLayout(new FlowLayout());
	panelTemp.add(fileHashButton = new JButton("Choose File"));
	panelTemp.add(objHashButton = new JButton("Choose Object"));
	panelTemp.add(hashChoice = new JComboBox());
	hashChoice.addItem("md5");
	hashChoice.addItem("sha1");
	panelTemp.add(createHashButton = new JButton("Create"));

	// control set: load identity
	panelIdentity = new JPanel();
	controls.addTab("Identity", panelIdentity);
	panelIdentity.setLayout(new BorderLayout());
	panelIdentity.add("Center",panelITop = new JPanel());
	panelITop.setLayout(new FlowLayout());
	panelFiles = new JPanel();
	panelITop.add(panelFiles);
	panelFiles.setLayout(new GridLayout(6,1));
	panelFiles.add(new JLabel("Public Key"));
	file1 = new JTextField(25);
	panelFiles.add(file1);
	chooseFile1 = new JButton("Choose File");
	panelFiles.add(chooseFile1);
	panelFiles.add(new JLabel("Private Key"));
	file2 = new JTextField(25);
	panelFiles.add(file2);
	chooseFile2 = new JButton("Choose File");
	panelFiles.add(chooseFile2);
	loadIdentity = new JButton("Load Identity");
	panelITop.add(loadIdentity);

	// control sets: prove and verify sequence
	panelProve  = new JPanel();
	controls.addTab("Prove", panelProve);
	panelProve.setLayout(new GridBagLayout());

	panelVerify = new JPanel();
	controls.addTab("Verify", panelVerify);
	panelVerify.setLayout(new GridBagLayout());

	setGBC(GridBagConstraints.CENTER, GridBagConstraints.BOTH);
	setGBC(0,0,1,1);

	panelProve.add(panelPTop = new JPanel(), gbc);
	panelPTop.setLayout(new GridBagLayout());

	panelVerify.add(panelVTop = new JPanel(), gbc);
	panelVTop.setLayout(new GridBagLayout());

	setGBC(0,1,1,0);

	panelVerify.add(panelVBottom = new JPanel(), gbc);
	panelVBottom.add(verifyButton = new JButton("Verify"));

	panelProve.add(panelPBottom = new JPanel(), gbc);
	panelPBottom.add(proveButton = new JButton("Prove"));

	setGBC(0,0,0,1);

	panelPTop.add(panelPControl = new JPanel(), gbc);
	panelPControl.setLayout(new GridLayout(4,1));

	panelVTop.add(panelVControl = new JPanel(), gbc);
	panelVControl.setLayout(new GridLayout(5,1));

	setGBC(1,0,1,1);

	panelPTop.add(panelPSelect = new JPanel(), gbc);
	panelPSelect.setLayout(new GridLayout(4,1));

	panelVTop.add(panelVSelect = new JPanel(), gbc);
	panelVSelect.setLayout(new GridLayout(5,1));

	// main prove panel

	panelPControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(issuerPButton = new JButton("Issuing Key"));
	panelTemp.add(new JLabel(":"));

	panelPSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(issuerPSelect = new JTextField(25));
	issuerPSelect.setEditable(false);

	panelPControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(new JLabel("Names:"));

	panelPSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(namesPSelect = new JTextField(25));

	panelPControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(new JLabel("Tag:"));

	panelPSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(tagPSelect = new JTextField(25));

	panelPControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(targetPButton = new JButton("Target Subject"));
	panelTemp.add(new JLabel(":"));

	panelPSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(targetPSelect = new JTextField(25));
	targetPSelect.setEditable(false);

	// main verify panel

	panelVControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(issuerVButton = new JButton("Issuing Key"));
	panelTemp.add(new JLabel(":"));

	panelVSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(issuerVSelect = new JTextField(25));
	issuerVSelect.setEditable(false);

	panelVControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(new JLabel("Names:"));

	panelVSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(namesVSelect = new JTextField(25));

	panelVControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(new JLabel("Tag:"));

	panelVSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(tagVSelect = new JTextField(25));

	panelVControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(targetVButton = new JButton("Target Subject"));
	panelTemp.add(new JLabel(":"));

	panelVSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(targetVSelect = new JTextField(25));
	targetVSelect.setEditable(false);

	panelVControl.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.RIGHT));
	panelTemp.add(sequenceVButton = new JButton("Proof Sequence"));
	panelTemp.add(new JLabel(":"));

	panelVSelect.add(panelTemp = new JPanel());
	panelTemp.setLayout(new FlowLayout(FlowLayout.LEFT));
	panelTemp.add(sequenceVSelect = new JTextField(25));
	sequenceVSelect.setEditable(false);

	// control set: create certs
	panelCerts = new JPanel();
	controls.addTab("Certs", panelCerts);
	panelCerts.setLayout(new GridLayout(3,1));
	panelCerts.add(topPanel = new JTabbedPane());
	panelCerts.add(panelCommon = new JPanel());

	// cert controls: create def
	panelNames = new JPanel();
	topPanel.addTab("Create Def", panelNames);
	panelNames.setLayout(new GridLayout(2,1));
	panelNames.add(buttonPanel = new JPanel());
	buttonPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
	buttonPanel.add(new JLabel("Create a new Name Definition Cert"));
	buttonPanel.add(createName = new JButton("Create Name"));
	panelNames.add(namePanel = new JPanel());
	namePanel.setLayout(new FlowLayout(FlowLayout.LEFT));
	namePanel.add(new JLabel("Name to Define"));
	namePanel.add(nameShown = new JTextField(25));

	// cert controls: create auth
	topPanel.addTab("Create Auth", panelAuths = new JPanel());
	panelAuths.setLayout(new GridLayout(2,1));
	panelAuths.add(authPanel = new JPanel());
	authPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
	authPanel.add(new JLabel("Create a new Authorization Cert"));
	authPanel.add(createAuth = new JButton("Create Auth"));
	panelAuths.add(authPanel2 = new JPanel());
	authPanel2.setLayout(new FlowLayout(FlowLayout.LEFT));
	authPanel2.add(new JLabel("(tag"));
	authPanel2.add(tagField = new JTextField(20));
	authPanel2.add(new JLabel(")"));
	authPanel2.add(propagate = new JCheckBox("Propagate"));

	// cert controls: common: select subject and validity
	panelCommon.setLayout(new GridLayout(4,1));
    panelCommon.add(panelTemp = new JPanel(new FlowLayout(FlowLayout.LEFT)));
	panelTemp.add(new JLabel("Subject:"));
	panelTemp.add(subjectShown = new SubjectField(25));
	panelTemp.add(subjectChoices = new JComboBox());
	subjectChoices.addItem("Double-Click");
	subjectChoices.addItem("Type a Name");
	panelCommon.add(new JLabel("Validity Dates: YYYY-MM-DD_hh:mm:ss"));
    panelCommon.add(panelTemp = new JPanel(new FlowLayout(FlowLayout.LEFT)));
    panelTemp.add(new JLabel("Not before:"));
	panelTemp.add(beforeField = new JTextField(19));
    panelCommon.add(panelTemp = new JPanel(new FlowLayout(FlowLayout.LEFT)));
	panelTemp.add(new JLabel("Not after:"));
	panelTemp.add(afterField = new JTextField(19));
	//}}

	this.pack();

	//{{INIT_MENUS
	mainMenuBar = new JMenuBar();
		
	mainMenuBar.add(fileMenu = new JMenu("File"));
	fileMenu.add(loadFileMenuItem = new JMenuItem("Load File"));
	fileMenu.add(saveSelectedMenuItem = new JMenuItem("Save Selected"));
	fileMenu.add(saveAllMenuItem = new JMenuItem("Save All"));
	fileMenu.add(clearAllMenuItem = new JMenuItem("Clear All"));
	fileMenu.add(quitMenuItem = new JMenuItem("Quit"));
		
	mainMenuBar.add(viewMenu = new JMenu("View"));
	viewMenu.add(curObjMenuItem = new JMenuItem("All"));
	viewMenu.add(viewAuthsMenuItem = new JMenuItem("Auths"));
	viewMenu.add(viewDefsMenuItem = new JMenuItem("Defs"));
	viewMenu.add(viewKeysMenuItem = new JMenuItem("Keys"));
	viewMenu.add(viewSigsMenuItem = new JMenuItem("Sigs"));
	viewMenu.add(viewSeqsMenuItem = new JMenuItem("Seqs"));
		
	mainMenuBar.add(objectMenu = new JMenu("Object"));
	objectMenu.add(signaturesMenuItem = new JMenuItem("Sign"));
	objectMenu.add(removeObjectMenuItem = new JMenuItem("Remove"));

	mainMenuBar.add(createMenu = new JMenu("Create"));
	createMenu.add(keysCreateMenuItem = new JMenuItem("Key Pair!"));
	createMenu.add(ohashCreateMenuItem = new JMenuItem("Object Hash!"));
	createMenu.add(closureCreateMenuItem = new JMenuItem("Closure!"));
	createMenu.add(proofCreateMenuItem = new JMenuItem("Proof!"));

	this.getRootPane().setJMenuBar(mainMenuBar);
	//}}


	//{{REGISTER_LISTENERS
	objectList.addListSelectionListener(new ListSelectionListener() {
		public void valueChanged(ListSelectionEvent e) {
          selectItem();
		}
      });
	objectList.addMouseListener(new MouseAdapter() {
		public void mouseClicked(MouseEvent e) {
          if (e.getClickCount() == 2) {
			// FIXME: remove this and replace with
			// a button for selection
			DisplayObject disp = selectItem();
			SDSIObject obj = disp.getObject();
			
			if (panelCerts.isVisible()
			    && subjectChoices.getSelectedItem()
			    .equals("Double-Click")) {
              if (obj instanceof Def) {
				obj = ((Def)obj).getName();
              } else if (!(obj instanceof Subject)) {
				statusText.setText("Select a Subject");
				return;
              }
              subjectShown.setSubject((Subject)obj);
              subjectShown.setText(disp.getName());
			}
          }
		}
      });
	curObjMenuItem.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent event)
        {
          listCurrentClass(SDSIObject.class);
        }
      });
	signaturesMenuItem.addActionListener(this);
	createObject.addActionListener(this);
	quitMenuItem.addActionListener(this);
	chooseFile1.addActionListener(this);
	chooseFile2.addActionListener(this);
	loadIdentity.addActionListener(this);
	subjectChoices.addItemListener(this);
	createName.addActionListener(this);
	createAuth.addActionListener(this);
	removeObjectMenuItem.addActionListener(this);
	viewAuthsMenuItem.addActionListener(this);
	viewDefsMenuItem.addActionListener(this);
	viewKeysMenuItem.addActionListener(this);
	viewSigsMenuItem.addActionListener(this);
	viewSeqsMenuItem.addActionListener(this);
	loadFileMenuItem.addActionListener(this);
	saveSelectedMenuItem.addActionListener(this);
	saveAllMenuItem.addActionListener(this);
	clearAllMenuItem.addActionListener(this);
	keysCreateMenuItem.addActionListener(this);
	ohashCreateMenuItem.addActionListener(this);
	closureCreateMenuItem.addActionListener(this);
	proofCreateMenuItem.addActionListener(this);
	issuerNButton.addActionListener(this);
	createNameButton.addActionListener(this);	
	addSeqButton.addActionListener(this);
	insertSeqButton.addActionListener(this);
	removeSeqButton.addActionListener(this);
	clearSeqButton.addActionListener(this);
	createSeqButton.addActionListener(this);
	loadSeqButton.addActionListener(this);
	addSubButton.addActionListener(this);
	insertSubButton.addActionListener(this);
	removeSubButton.addActionListener(this);
	clearSubButton.addActionListener(this);
	createSubButton.addActionListener(this);
	loadSubButton.addActionListener(this);
	fileHashButton.addActionListener(this);
	objHashButton.addActionListener(this);
	createHashButton.addActionListener(this);
	issuerVButton.addActionListener(this);
	targetVButton.addActionListener(this);
	sequenceVButton.addActionListener(this);
	verifyButton.addActionListener(this);
	issuerPButton.addActionListener(this);
	targetPButton.addActionListener(this);
	proveButton.addActionListener(this);
	//}}
  }


  //{{DECLARE_CONTROLS
  JPanel panelTemp; // used for layout only
  JLabel statusLabel;
  JTextField statusText;
  JLabel identityLabel;
  JTextField identityText;

  JPanel main;
  JPanel rightPanel;
  JPanel displayPanel;
  JTextArea display;
  JPanel panelObjList;
  JList objectList;

  JTabbedPane controls;

  JPanel panelObjects;
  JTextArea newObject;
  JButton createObject;

  JPanel panelName;
  JButton issuerNButton;
  JTextField issuerNSelect;
  JTextField nameText;
  JButton createNameButton;

  JPanel panelSequence;
  JList sequenceList;
  JButton addSeqButton;
  JButton insertSeqButton;
  JButton removeSeqButton;
  JButton clearSeqButton;
  JButton createSeqButton;
  JButton loadSeqButton;
    
  JPanel panelThreshold;
  JTextField thresholdText;
  JList subjectsList;
  JButton addSubButton;
  JButton insertSubButton;
  JButton removeSubButton;
  JButton clearSubButton;
  JButton createSubButton;
  JButton loadSubButton;
    
  JPanel panelHash;
  JTextField hashItemName;
  JButton fileHashButton;
  JButton objHashButton;
  JComboBox hashChoice;
  JButton createHashButton;

  JPanel panelIdentity;
  JPanel panelITop;
  JPanel panelFiles;
  JButton chooseFile1;
  JButton chooseFile2;
  JTextField file1;
  JTextField file2;
  JButton loadIdentity;

  JPanel panelProve;
  JPanel panelPTop;
  JPanel panelPBottom;
  JPanel panelPControl;
  JPanel panelPSelect;
  JButton proveButton;

  JButton issuerPButton;
  JTextField issuerPSelect;
  JTextField namesPSelect;
  JTextField tagPSelect;
  JButton targetPButton;
  JTextField targetPSelect;

  JPanel panelVerify;
  JPanel panelVTop;
  JPanel panelVBottom;
  JPanel panelVControl;
  JPanel panelVSelect;
  JButton verifyButton;

  JButton issuerVButton;
  JTextField issuerVSelect;
  JTextField namesVSelect;
  JTextField tagVSelect;
  JButton targetVButton;
  JTextField targetVSelect;
  JButton sequenceVButton;
  JTextField sequenceVSelect;

  JPanel panelCerts;
  JTabbedPane topPanel;

  JPanel panelNames;
  JPanel buttonPanel;
  JButton createName;
  JPanel namePanel;
  JTextField nameShown;

  JPanel panelAuths;
  JButton createAuth;
  JPanel authPanel;
  JPanel authPanel2;
  JTextField tagField;
  JCheckBox propagate;

  JPanel panelCommon;
  JComboBox subjectChoices;
  JTextField beforeField;
  JTextField afterField;
  //}}
    
  SubjectField subjectShown;

  //{{DECLARE_MENUS
  JMenuBar mainMenuBar;
  JMenu fileMenu;
  JMenuItem loadFileMenuItem;
  JMenuItem saveSelectedMenuItem;
  JMenuItem saveAllMenuItem;
  JMenuItem clearAllMenuItem;
  JMenuItem quitMenuItem;

  JMenu viewMenu;
  JMenuItem curObjMenuItem;
  JMenuItem viewDefsMenuItem;
  JMenuItem viewAuthsMenuItem;
  JMenuItem viewKeysMenuItem;
  JMenuItem viewSigsMenuItem;
  JMenuItem viewSeqsMenuItem;

  JMenu createMenu;
  JMenuItem keysCreateMenuItem;
  JMenuItem ohashCreateMenuItem;
  JMenuItem closureCreateMenuItem;
  JMenuItem proofCreateMenuItem;

  JMenu objectMenu;
  JMenuItem signaturesMenuItem;
  JMenuItem removeObjectMenuItem;
  //}}

  class SubjectField extends JTextField {
	private Subject subj;
	SubjectField(int k) {
      super(k);
      subj = null;
      this.setText("Please select a subject");
      this.setEditable(false);
	}
	Subject getSubject() throws SDSIException {
      if (subjectChoices.getSelectedItem().equals("Type a Name")) {
		try {
          if (this.getText().equals(""))
			throw new SDSIException("Need a name");
          StringBuffer name = new StringBuffer("(name ");
          name.append(currentPublicKey.toString() + " ");
          name.append(this.getText() + ")");
          return (Name)processStream
			(new ByteArrayInputStream(name.toString()
                                      .getBytes("8859_1")),
			 false);
		}
		catch (UnsupportedEncodingException e) {
          return null; //won't happen
		}
      } else {
		return subj;
      }
	}
	void setSubject(Subject s) {
      subj = s;
      // UNCOMMENT TO HASH ALL CERT SUBJECTS
      //  	    if (subj instanceof SDSIPublicKey) {
      //  		try {
      //  		    // hash the subject key
      //  		    subj = ((SDSIPublicKey)subj).getHash();
      //  		} catch (NoSuchAlgorithmException e) {
      //  		    throw new Error
      //  			("No Provider (NoSuchAlgorithmException): "
      //  			 +e.getMessage());
      //  		}
      //  	    }
	}
  }

  public void actionPerformed(ActionEvent event)
  {
	statusText.setText("");
	Object object = event.getSource();
	if (object == chooseFile1)
      chooseFile1_Action(event);
	else if (object == chooseFile2)
      chooseFile2_Action(event);
	else if (object == loadIdentity)
      loadIdentity_Action(event);
	else if (object == quitMenuItem)
      quitMenuItem_Action(event);
	else if (object == signaturesMenuItem)
      signaturesMenuItem_Action(event);
	else if (object == createObject)
      createObject_Action(event);
	else if (object == createName)
      createName_Action(event);
	else if (object == createAuth)
      createAuth_Action(event);
	else if (object == removeObjectMenuItem)
      removeObjectMenuItem_Action(event);
	else if (object == viewAuthsMenuItem)
      viewAuthsMenuItem_Action(event);
	else if (object == viewDefsMenuItem)
      viewDefsMenuItem_Action(event);
	else if (object == viewKeysMenuItem)
      viewKeysMenuItem_Action(event);
	else if (object == viewSigsMenuItem)
      viewSigsMenuItem_Action(event);
	else if (object == viewSeqsMenuItem)
      viewSeqsMenuItem_Action(event);
	else if (object == loadFileMenuItem)
      loadFileMenuItem_Action(event);
	else if (object == saveSelectedMenuItem)
      saveSelectedMenuItem_Action(event);
	else if (object == saveAllMenuItem)
      saveAllMenuItem_Action(event);
	else if (object == clearAllMenuItem)
      clearAllMenuItem_Action(event);
	else if (object == keysCreateMenuItem)
      keysCreateMenuItem_Action(event);
	else if (object == ohashCreateMenuItem)
      ohashCreateMenuItem_Action(event);
	else if (object == closureCreateMenuItem)
      closureCreateMenuItem_Action(event);
	else if (object == proofCreateMenuItem)
      proofCreateMenuItem_Action(event);
	else if (object == issuerNButton)
      issuerNButton_Action(event);
	else if (object == createNameButton)
      createNameButton_Action(event);

	else if (object == addSeqButton)
      addSeqButton_Action(event);
	else if (object == insertSeqButton)
      insertSeqButton_Action(event);
	else if (object == removeSeqButton)
      removeSeqButton_Action(event);
	else if (object == clearSeqButton)
      clearSeqButton_Action(event);
	else if (object == createSeqButton)
      createSeqButton_Action(event);
	else if (object == loadSeqButton)
      loadSeqButton_Action(event);

	else if (object == addSubButton)
      addSubButton_Action(event);
	else if (object == insertSubButton)
      insertSubButton_Action(event);
	else if (object == removeSubButton)
      removeSubButton_Action(event);
	else if (object == clearSubButton)
      clearSubButton_Action(event);
	else if (object == createSubButton)
      createSubButton_Action(event);
	else if (object == loadSubButton)
      loadSubButton_Action(event);

	else if (object == fileHashButton)
      fileHashButton_Action(event);
	else if (object == objHashButton)
      objHashButton_Action(event);
	else if (object == createHashButton)
      createHashButton_Action(event);
	else if (object == issuerVButton)
      issuerVButton_Action(event);
	else if (object == targetVButton)
      targetVButton_Action(event);
	else if (object == sequenceVButton)
      sequenceVButton_Action(event);
	else if (object == verifyButton)
      verifyButton_Action(event);
	else if (object == issuerPButton)
      issuerPButton_Action(event);
	else if (object == targetPButton)
      targetPButton_Action(event);
	else if (object == proveButton)
      proveButton_Action(event);
  }

  public void itemStateChanged(ItemEvent event)
  {
	Object object = event.getSource();
	if (object == subjectChoices)
      subjectChoices_ItemStateChanged(event);
  }

  // listCurrentClass refreshes the list with items of the specified class
  // and caches the last class used in currentListClass.
  Class currentListClass = SDSIObject.class;
  void listCurrentClass(Class c) {
	if (c != null) currentListClass = c;

	currentObjects.clear();
	Iterator objects = cache.getObjects().iterator();
	while (objects.hasNext()) {
      SDSIObject object = (SDSIObject)objects.next();
      if (currentListClass.isInstance(object))
		addItem(new DisplayObject(object));
	}
  }

  // insertion sort :P
  private void addItem(DisplayObject disp)
  {
	Enumeration items = currentObjects.elements();
	int index = 0;
	while (items.hasMoreElements()) {
      DisplayObject item = (DisplayObject)items.nextElement();
      if (disp.compareTo(item) < 0) {
		currentObjects.insertElementAt(disp, index);
		return;
      }
      index++;
	}
	currentObjects.addElement(disp);
  }
    
  private SDSIObject processStream(InputStream is, boolean add) {
	try {
      if (is == null) throw new IOException("Null Stream");
      SDSIObject object = SDSIObject.readFrom(is, currentPublicKey);
      if (add && (object != null) && (cache != null)) {
		cache.add(object);
      }
      return object;
	} catch (IOException e) {
      reportError(e);
      return null;
	} catch (SDSIException e) {
      reportError(e);
      return null;
	}
  }
    
  private void displayObject(SDSIObject object)
  {
	if (object == null) {
      statusText.setText("");
      display.setText("");
      return;
	}
	statusText.setText(object.getType());
	display.setText(object.toReadableString(display.getColumns()));
  }

  private void refreshDisplay()
  {
	displayObject(currentObject);
  }

  private void refreshList()
  {
	listCurrentClass(null);
  }

  private void reportError(String title, String message)
  {
	new ErrorDialog(this, title, message).setVisible(true);
  }

  private void reportError(String title, Throwable t)
  {
    StringWriter s = new StringWriter();
    t.printStackTrace(new PrintWriter(s, true));
    reportError(title, s.toString());
  }
  
  private void reportError(Throwable t)
  {
    reportError(t.getClass().getName(), t);
  }

  private void addObject(SDSIObject object)
  {
	cache.add(object);
	refreshList();
  }

  private void selectObject(SDSIObject object)
  {
	currentObject = object;
	refreshDisplay();
  }

  private void addAndSelectObject(SDSIObject object)
  {
	addObject(object);
    for (int i = 0; i < currentObjects.size(); i++) {
      if (((DisplayObject)currentObjects.get(i))
          .getObject().equals(object)) {
        objectList.setSelectedIndex(i);
        break;
      }
    }
	selectObject(object);
  }

  private DisplayObject selectItem()
  {
	DisplayObject disp = getSelected();
	selectObject(disp.getObject());
	return disp;
  }

  private DisplayObject nullDisplayObject =
	new DisplayObject(null); // XXX needed?
    
  private DisplayObject getItem(int index)
  {
	if ((index < 0) || (index >= currentObjects.size()))
      return nullDisplayObject;
	return (DisplayObject)currentObjects.get(index);
  }

  private DisplayObject getSelected()
  {
	return getItem(objectList.getSelectedIndex());
  }

  void signaturesMenuItem_Action(ActionEvent event)
  {
	if (currentPublicKey == null) {
      reportError("Identity Error",
                  "Set identity for signing");
      return;
    }
    if (currentObject == null) {
      reportError("Select an Object",
                  "Select an object to sign");
      return;
    }
    ConfirmDialog query = new ConfirmDialog(this, "Sign Object?");
    query.setVisible(true);
    if (query.answer())
      signObject();
  }

  void createObject_Action(ActionEvent event)
  {
	String newString = newObject.getText();
	int numObjects = 0;
	try {
      ByteArrayInputStream input =
		new ByteArrayInputStream(newString.getBytes("8859_1"));
      while (processStream(input, true) != null)
		numObjects++;
      statusText.setText(numObjects + " S-expression"
                         + ((numObjects == 1)?"":"s")
                         + " parsed");
      refreshList();
      //newObject.setText("");
	}
  	catch (UnsupportedEncodingException e) {
      reportError(e);
  	}
  }

  void quitMenuItem_Action(ActionEvent event)
  {
	this.dispose();
	System.exit(0);
  }

  File currentDirectory = null;
  String getFilename(String title, int type)
  {
	JFileChooser chooser = new JFileChooser(currentDirectory);
	chooser.setDialogTitle(title);
	chooser.setDialogType(type);
	int result = chooser.showDialog(this, null);
	if (result == JFileChooser.APPROVE_OPTION) {
      currentDirectory = chooser.getCurrentDirectory();
      try {
		return chooser.getSelectedFile().getCanonicalPath();
      } catch (IOException e) {
		reportError(e);
		return null;
      }
	}
	return null;
  }

  void chooseFile1_Action(ActionEvent event)
  {
	String filename = getFilename("Choose Public Key",
                                  JFileChooser.OPEN_DIALOG);
	if (filename != null)
      file1.setText(filename);
  }

  void chooseFile2_Action(ActionEvent event)
  {
	String filename = getFilename("Choose Private Key",
                                  JFileChooser.OPEN_DIALOG);
	if (filename != null)
      file2.setText(filename);
  }

  void loadIdentity_Action(ActionEvent event)
  {
	String f1 = file1.getText();
	String f2 = file2.getText();
	if (f1.equals("") || f2.equals("")) {
      reportError("File Error", "Please specify both "
                  + "a Public Key and a Private Key");
      return;
	}
	try {
      // read public key from file
      FileInputStream fis = new FileInputStream(f1);
      SDSIObject o = processStream(fis, false);
      fis.close();
      if (!(o instanceof SDSIPublicKey)) {
		reportError("SDSIObject Error",
                    "Public key file must contain a public key");
		//take out the object used to be here
		return;
      }
      SDSIPublicKey pub_key = (SDSIPublicKey)o;
      addObject(pub_key);

      // read private key from file
      fis = new FileInputStream(f2);
      o = processStream(fis, false);
      fis.close();
      if (!(o instanceof SDSIPrivateKey)) {
		reportError("SDSIObject Error",
                    "Private key file must contain a private key");
		return;
      }
      SDSIPrivateKey priv_key = (SDSIPrivateKey)o;

      // test public and private keys match
      Signature sign = Signature.getInstance("MD5/RSA/PKCS#1");
      Signature verify = Signature.getInstance("MD5/RSA/PKCS#1");
      sign.initSign(priv_key);
      verify.initVerify(pub_key);
      byte[] test = new String("Babbaloo Babbaloo Hey How About You?")
		.getBytes();
      sign.update(test);
      verify.update(test);
      if (verify.verify(sign.sign())) {
		currentPublicKey = pub_key;
		currentPrivateKey = priv_key;
      } else {
		reportError("Identity Error",
                    "Public Key and Private Key do not match");
      }
	    
      // add public key to cache
      statusText.setText("Identity set");
      identityText.setText(cache.getDisplayName(null, currentPublicKey));
      refreshList();	    
	} catch (IOException e) {
      reportError(e);
	} catch (NoSuchAlgorithmException e) {
      reportError(e);
	} catch (Exception e){
      reportError(e);
	}
  }

  void subjectChoices_ItemStateChanged(ItemEvent event)
  {
	String choiceMethod = (String)event.getItem();
	if (choiceMethod.equals("Type a Name")) {
      subjectShown.setText("");
      subjectShown.setEditable(true);
	}
	else {
      subjectShown.setText("Please Select a Subject");
      subjectShown.setEditable(false);
	}
  }

  void createName_Action(ActionEvent event)
  {
	// TODO: This code is generating a null pointer sometimes
	try {
      Subject s = subjectShown.getSubject();
      Validity v = new Validity();
      String name = nameShown.getText();
      if (name.equals("")) {
		reportError("Add a name","Must add at least one name");
		return;
      }
      v.updateStrings(beforeField.getText(),afterField.getText());
      Def newDef = new Def(currentPublicKey/*.getHash()*/,name,s,v);
      addAndSelectObject(newDef);
      signaturesMenuItem_Action(null);
	} catch (SDSIException e) {
      reportError(e);
      return;
	} catch (java.text.ParseException e) {
      reportError(e);
      return;
	}/* catch (NoSuchAlgorithmException e) {
	    reportError(e);
	    return;
	    }*/
	// UNCOMMENT PREVIOUS SECTIONS TO HASH CERT ISSUERS
  }

  Tag processTag() {
   	Sexp[] majorTagParts = new Sexp[2];
	majorTagParts[0] = new SexpString("tag");
	try {
      String newTag = tagField.getText();
      PushbackInputStream p = new PushbackInputStream
		(new ByteArrayInputStream(newTag.getBytes("8859_1")));
      majorTagParts[1] = Sexp.parse(p);
      SexpList tagList = new SexpList(majorTagParts);
      return new Tag(tagList);
	}
	catch (UnsupportedEncodingException e) {
      reportError(e);
  	}
  	catch (SexpParseException e) {
      reportError(e);
  	}
	catch (SexpException e) {
      reportError(e);
	}
	return null;
  }

  void createAuth_Action(ActionEvent event)
  {
	try {
      Subject s = subjectShown.getSubject();
      Validity v = new Validity();
      boolean prop = propagate.isSelected();
      Tag t = processTag();
      if (t == null || s == null) {
		reportError("Insufficient Information",
                    "Please specify both a subject and a tag");
		return;
      }
      v.updateStrings(beforeField.getText(),afterField.getText());
      Auth newAuth = new Auth(currentPublicKey,s,t,prop,v);
      addAndSelectObject(newAuth);
      signaturesMenuItem_Action(null);
	}
	catch (SDSIException e) {
      reportError(e);
      return;
	}
	catch (java.text.ParseException e) {
      reportError(e);
      return;
	}
  }

  void signObject()
  {
	try {
      SDSISignature newSig = new SDSISignature(currentObject,
                                               currentPrivateKey,
                                               currentPublicKey);
      addAndSelectObject(newSig);
	} catch (SignatureException e) {
      reportError(e);
	}
  }

  void removeObjectMenuItem_Action(ActionEvent event)
  {
	if (currentObject == null) {
      statusText.setText("Select an Object to remove");
      return;
	}
	ConfirmDialog query = new ConfirmDialog(this, "Remove Object?");
	query.setVisible(true);
	if (query.answer() == true) {
      cache.remove(currentObject);
      statusText.setText(currentObject.getType() + " removed");
      currentObject = null;
      refreshList();
	}   
  }

  void viewAuthsMenuItem_Action(ActionEvent event)
  {
	listCurrentClass(Auth.class);
  }

  void viewDefsMenuItem_Action(ActionEvent event)
  {
	listCurrentClass(Def.class);
  }

  void viewKeysMenuItem_Action(ActionEvent event)
  {
	listCurrentClass(SDSIPublicKey.class);
  }

  void viewSigsMenuItem_Action(ActionEvent event)
  {
	listCurrentClass(SDSISignature.class);
  }

  void viewSeqsMenuItem_Action(ActionEvent event)
  {
	listCurrentClass(Sequence.class);
  }

  private boolean checkCache()
  {
	if (cache == null) {
      reportError("Internal Error", "Cache is NULL");
      return false;
	}
	return true;
  }

  void loadFileMenuItem_Action(ActionEvent event)
  {
	String file = getFilename("Read in S-Expressions",
                              JFileChooser.OPEN_DIALOG);
   	if (file == null) return;
	int countSexps = 0;
	try {
      FileInputStream fis = new FileInputStream(file);
      while (processStream(fis, true) != null)
		countSexps++;
      fis.close();
	} catch (IOException e) {
      reportError(e);
      return;
	}
	statusText.setText(countSexps + " S-expression"
                       + ((countSexps == 1)?"":"s")
                       + " loaded");
	refreshList();
  }
    
  private void saveObject(SDSIObject object, String name)
  {
	// if object == null, panic
	String file = getFilename("Save "+name,
                              JFileChooser.SAVE_DIALOG);
	if (file == null) return;
	if (!checkCache()) return;
	try {
      FileOutputStream fos = new FileOutputStream(file);
      object.writeReadable(fos);
      fos.close();
	} catch (IOException e) {
      reportError(e);
      return;
	}
	statusText.setText(name+" saved");
  }

  void saveSelectedMenuItem_Action(ActionEvent event)
  {
	DisplayObject disp = getSelected();
	if (disp.getObject() == null) {
      statusText.setText("Select an Object");
      return;
	}
	saveObject(disp.getObject(), disp.getName());
  }

  void saveAllMenuItem_Action(ActionEvent event)
  {
	String file = getFilename("Save S-Expressions",
                              JFileChooser.SAVE_DIALOG);
	if (file == null) return;
	if (!checkCache()) return;
	try {
      FileOutputStream fos = new FileOutputStream(file);
      cache.writeTo(fos);
      fos.close();
	} catch (IOException e) {
      reportError(e);
      return;
	}
	statusText.setText("S-Expressions saved");
  }

  void clearAllMenuItem_Action(ActionEvent event)
  {
	initializeCache();
	refreshDisplay();
	listCurrentClass(SDSIObject.class); // reset displayed type
  }


  void ohashCreateMenuItem_Action(ActionEvent event)
  {
	SDSIObject object = getSelected().getObject();
	if (!(object instanceof Hash)) {
      statusText.setText("Select a Hash");
      return;
	}
	try {
      ObjectHash ohash = new ObjectHash((Hash)object);
      addAndSelectObject(ohash);
	} catch (SexpParseException e) {
      reportError(e);
      return;
	}
  }

  void keysCreateMenuItem_Action(ActionEvent event)
  {
    try {
      statusText.setText("Please Wait");
      KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA");
      kpg.initialize(1024,new SecureRandom());
      KeyPair kp = kpg.generateKeyPair();
      PublicKey pubKey = kp.getPublic();
      PrivateKey privKey = kp.getPrivate();
      SDSIRSAPrivateKey sdsiPrivKey =
		new SDSIRSAPrivateKey((RSAPrivateKey)privKey);
      SDSIRSAPublicKey sdsiPubKey =
		new SDSIRSAPublicKey((RSAPublicKey)pubKey);
      displayObject(sdsiPrivKey);
      saveObject(sdsiPrivKey, "Private Key");
      displayObject(sdsiPubKey);
      saveObject(sdsiPubKey, "Public Key");
      addAndSelectObject(sdsiPubKey);
    } catch (NoSuchAlgorithmException e) {
      reportError(e);
    }
  }

  SDSIPublicKey nameIssuerKey = null;
  void issuerNButton_Action(java.awt.event.ActionEvent event)
  {
    DisplayObject disp = getSelected();
    if (disp.getObject() instanceof SDSIPublicKey) {
      nameIssuerKey = (SDSIPublicKey)disp.getObject();
      issuerNSelect.setText(disp.getName());
    } else {
      statusText.setText("Select a Public Key");
    }
  }   

  String[] getTokens(String text)
  {
	StringTokenizer tokens = new StringTokenizer(text);
	if (tokens.countTokens() <= 0) return null;
	String[] names = new String[tokens.countTokens()];
	for (int i = 0; i < names.length; i++)
      names[i] = tokens.nextToken();
	return names;
  }

  void createNameButton_Action(java.awt.event.ActionEvent event)
  {
	if (nameIssuerKey == null) {
      statusText.setText("Select an Issuer");
      return;
	}

	String[] names = getTokens(nameText.getText());
	if (names == null) {
      statusText.setText("Type Names");
      return;
	}

	try {
      addAndSelectObject(new Name(nameIssuerKey, names));
	} catch (SDSIException e) {
      reportError(e);
	}
  }

  boolean checkType(Object obj, Class type, String typeName)
  {
	if (type.isInstance(obj)) return true;
	reportError("Not a "+typeName,
                "Selected Item must be a "+typeName);
	return false;
  }

  void addSelected(DefaultListModel listModel,
                   Class listType,
                   String typeName)
  {
	DisplayObject disp = getSelected();
	SDSIObject obj = disp.getObject();
	if (!checkType(obj, listType, typeName)) return;
	if (obj == null) return; // XXX needed?
	listModel.addElement(disp);
  }

  void addSeqButton_Action(java.awt.event.ActionEvent event)
  {
	addSelected(currentSequence, SequenceItem.class, "Sequence Item");
  }

  void addSubButton_Action(java.awt.event.ActionEvent event)
  {
	addSelected(currentSubjects, Subject.class, "Subject");
  }

  void insertSelected(JList list,
                      DefaultListModel listModel,
                      Class listType,
                      String typeName)
  {
	DisplayObject disp = getSelected();
	SDSIObject obj = disp.getObject();
	if (!checkType(obj, listType, typeName)) return;
	if (obj == null) return; // XXX needed?
	int index = list.getSelectedIndex();
	if (index < 0 || index > listModel.size()) {
      // XXX can this happen?
      listModel.addElement(disp);
	} else {
      listModel.add(index, disp);
	}
  }

  void insertSeqButton_Action(java.awt.event.ActionEvent event)
  {
	insertSelected(sequenceList, currentSequence,
                   SequenceItem.class, "Sequence Item");
  }

  void insertSubButton_Action(java.awt.event.ActionEvent event)
  {
	insertSelected(subjectsList, currentSubjects,
                   Subject.class, "Subject");
  }

  void removeSelected(JList list, DefaultListModel listModel)
  {
	int index = list.getSelectedIndex();
	if (index < 0) {
      reportError("Select an Item","Select an Item to remove");
      return;
	}
	listModel.remove(index);
  }

  void removeSeqButton_Action(java.awt.event.ActionEvent event)
  {
	removeSelected(sequenceList, currentSequence);
  }

  void removeSubButton_Action(java.awt.event.ActionEvent event)
  {
	removeSelected(subjectsList, currentSubjects);
  }

  void clearSeqButton_Action(java.awt.event.ActionEvent event)
  {
	currentSequence.clear();
  }

  void clearSubButton_Action(java.awt.event.ActionEvent event)
  {
	currentSubjects.clear();
	thresholdText.setText("");
  }

  void createSeqButton_Action(java.awt.event.ActionEvent event)
  {
	SequenceItem[] members = new SequenceItem[currentSequence.size()];
	for (int i = 0; i < members.length; i++)
      members[i] = (SequenceItem)
		((DisplayObject)currentSequence.get(i)).getObject();

	try {
      Sequence sequence = Sequence.createFrom(members);
      addAndSelectObject(sequence);
	} catch (SexpParseException e) {
      reportError(e);
      return;
	}
  }

  void createSubButton_Action(java.awt.event.ActionEvent event)
  {
	Subject[] subjects = new Subject[currentSubjects.size()];
	for (int i = 0; i < subjects.length; i++)
      subjects[i] = (Subject)
		((DisplayObject)currentSubjects.get(i)).getObject();

	int threshold;
	try {
      threshold = Integer.parseInt(thresholdText.getText());
	} catch (NumberFormatException e) {
      reportError("Enter a Threshold",
                  "Enter a number for the threshold");
      return;
	}

	try {
      Threshold thresh = Threshold.create(threshold, subjects);
      addAndSelectObject(thresh);
	} catch (SDSIException e) {
      reportError("Sexp Parse Error",
                  "Error Parsing Threshold: "
                  + e.getMessage());
      return;
	}
  }

  void loadSeqButton_Action(java.awt.event.ActionEvent event)
  {
	DisplayObject disp = getSelected();
	if (!(disp.getObject() instanceof Sequence)) {
      statusText.setText("Select a Sequence");
      return;
	}

	currentSequence.clear();
	Sequence sequence = (Sequence)disp.getObject();
	SDSIObject[] members = sequence.members();
	for (int i = 0; i < members.length; i++) {
      currentSequence.addElement(new DisplayObject(members[i]));
	}
  }

  void loadSubButton_Action(java.awt.event.ActionEvent event)
  {
	DisplayObject disp = getSelected();
	if (!(disp.getObject() instanceof Threshold)) {
      statusText.setText("Select a Threshold");
      return;
	}

	currentSubjects.clear();
	Threshold thresh = (Threshold)disp.getObject();
	Subject[] subjects = thresh.getSubjects();
	for (int i = 0; i < subjects.length; i++) {
      currentSubjects.addElement(new DisplayObject
		((SDSIObject)subjects[i])); // XXX cast to SDSIObject safe?
	}
	thresholdText.setText(Integer.toString(thresh.getThreshold()));
  }

  SDSIObject objectToHash = null;
  void fileHashButton_Action(java.awt.event.ActionEvent event)
  {
	String filename = getFilename("Select File to Hash",
                                  JFileChooser.OPEN_DIALOG);
	if (filename == null) return;
	hashItemName.setText(filename);
	objectToHash = null;
  }

  void objHashButton_Action(java.awt.event.ActionEvent event)
  {
	DisplayObject disp = getSelected();
	objectToHash = disp.getObject();
	if (objectToHash == null) {
      statusText.setText("Select Object to Hash");
      return;
	}
	hashItemName.setText(disp.getName());
  }

  Hash hashFile()
  {
	try {
      File file = new File(hashItemName.getText());
      if (!file.exists())
		throw new FileNotFoundException("File not Found");

      byte[] data = new byte[(int)file.length()];
      FileInputStream input = new FileInputStream(file);
      input.read(data);
      MessageDigest digest =
		MessageDigest.getInstance
		((String)hashChoice.getSelectedItem());
      return new Hash((String)hashChoice.getSelectedItem(),
                      digest.digest(data));
	} catch (IOException e) {
      reportError(e);
	} catch (NoSuchAlgorithmException e) {
      reportError(e);
	} catch (SexpParseException e) {
      reportError(e);	    
	}
	return null;
  }

  Hash hashObject()
  {
	try {
      return new Hash((String)hashChoice.getSelectedItem(),
                      objectToHash);
	} catch (NoSuchAlgorithmException e) {
      reportError(e);
	}
	return null;
  }

  void createHashButton_Action(java.awt.event.ActionEvent event)
  {
	Hash hash;
	if (objectToHash == null)
      hash = hashFile();
	else
      hash = hashObject();

	if (hash == null) return;
	
	addAndSelectObject(hash);
  }

  private void selectIssuerKey()
  {
	DisplayObject disp = getSelected();

	if (disp.getObject() instanceof SDSIPublicKey) {
      issuerKey = (SDSIPublicKey)disp.getObject();
      issuerPSelect.setText(disp.getName());
      issuerVSelect.setText(disp.getName());
	} else {
      statusText.setText("Select a Public Key");
	}
  }

  void issuerPButton_Action(java.awt.event.ActionEvent event)
  {
	selectIssuerKey();
  }

  void issuerVButton_Action(java.awt.event.ActionEvent event)
  {
	selectIssuerKey();
  }

  private void selectTargetSubject()
  {
	DisplayObject disp = getSelected();
	
	if (disp.getObject() instanceof Subject) {
      targetSubject = (Subject)disp.getObject();
      targetPSelect.setText(disp.getName());
      targetVSelect.setText(disp.getName());
	} else {
      statusText.setText("Select a Subject");
	}
  }

  void targetPButton_Action(java.awt.event.ActionEvent event)
  {
	selectTargetSubject();
  }

  void targetVButton_Action(java.awt.event.ActionEvent event)
  {
	selectTargetSubject();
  }

  void sequenceVButton_Action(java.awt.event.ActionEvent event)
  {
	DisplayObject disp = getSelected();

	if (disp.getObject() instanceof Sequence) {
      proofSequence = (Sequence)disp.getObject();
      sequenceVSelect.setText(disp.getName());
	} else {
      statusText.setText("Select a Sequence");
	}
  }

  SDSIPublicKey issuerKey = null;
  Subject targetSubject = null;
  Sequence proofSequence  = null;

  void verifyButton_Action(java.awt.event.ActionEvent event)
  {
	String[] tags = getTokens(tagVSelect.getText());
	String[] names = getTokens(namesVSelect.getText());
	if ((issuerKey == null)
	    || (targetSubject == null)
	    || (proofSequence == null)) {
      reportError("Choose Parameters",
                  "Choose an Issuer Key, a Target Key,"
                  +" and a Proof Sequence");
      return;
	}
	try {
      Verifier verifier = new Verifier();
      boolean result;
      if (tags == null){
		if (names == null) {
          statusText.setText("Type Names or Tag or Both");
          return;
		}
		result = verifier.verifyDef(proofSequence,
                                    issuerKey,
                                    names,
                                    targetSubject);
      } else {
		result = verifier.verifyAuth(proofSequence,
                                     issuerKey,
                                     names,   // null ok
                                     new Tag(tags[0]),
                                     targetSubject);
		// FIXME: only supports bytestring tags
      }
      statusText.setText("Verification "+(result?"Passed":"Failed"));
	} catch (SDSIException e) {
      reportError(e);
	}
  }

  void proveButton_Action(java.awt.event.ActionEvent event)
  {
	String[] tags = getTokens(tagPSelect.getText());
	String[] names = getTokens(namesPSelect.getText());
	if ((issuerKey == null) || (targetSubject == null)) {
      reportError("Choose Parameters",
                  "Choose an Issuer Key and a Target Key");
      return;
	}
	try {
      SearchProver prover = new SearchProver();
      Sequence sequence = null;

      if (tags == null) {
		if (names == null) {
          statusText.setText("Type Names or Tag or Both");
          return;
		}
		sequence = prover.proveDef(cache,
                                   issuerKey,
                                   names,
                                   targetSubject);
      } else {
		sequence = prover.proveAuth(cache,
                                    issuerKey,
                                    names, // null ok
                                    new Tag(tags[0]),
                                    targetSubject);
		// FIXME: only supports bytestring tags
      }
	
      addProofSequence(sequence);

      // copy tag and names to verifier panel
      tagVSelect.setText(tagPSelect.getText());
      namesVSelect.setText(namesPSelect.getText());

	} catch (SDSIException e) {
      reportError(e);
      return;
	}
  }

  void closureCreateMenuItem_Action(java.awt.event.ActionEvent event)
  {
	cache.fullClosure();
	refreshList();
  }

  void addProofSequence(Sequence sequence)
  {
	if (sequence == null) {
      statusText.setText("Proof Failed");
      return;
	}

	statusText.setText("Proof Succeeded");
	addAndSelectObject(sequence);
  }

  void proofCreateMenuItem_Action(java.awt.event.ActionEvent event)
  {
	SDSIObject object = getSelected().getObject();

	if (!(object instanceof Cert)) {
      statusText.setText("Select a Cert");
      return;
	}

	Cert cert = (Cert)object;
	ClosureProver prover = new ClosureProver();
	Sequence sequence = null;
	try {
      sequence = prover.prove(cache, cert);
	} catch (SDSIException e) {
      reportError(e);
      return;
	}

	addProofSequence(sequence);
  }
}
